from hashlib import sha256
import time
from typing import List

import pytest

from tests.utils import MockNode

from mimblewimble.entity import OutputDataEntity
from mimblewimble.models.transaction import EOutputStatus

from mimblewimble.wallet import Wallet
from mimblewimble.wallet import WalletStorageInMemory, PersistentWallet, NodeAccess

from mimblewimble.models.slatepack.address import SlatepackAddress
from mimblewimble.models.slatepack.metadata import SlatepackVersion
from mimblewimble.models.slatepack.metadata import SlatepackMetadata
from mimblewimble.models.slatepack.message import SlatepackMessage, EMode

def test_srs_flow_slatepacks():
    # prepare wallets
    alice_path = 'm/0/1/0'
    alice_wallet = Wallet.initialize()
    alice_address = alice_wallet.getSlatepackAddress()

    bob_path = 'm/0/1/0'
    bob_wallet = Wallet.initialize()
    bob_address = bob_wallet.getSlatepackAddress()

    # prepare transaction data
    coinbase_amount = 60000000000
    kernel, output = alice_wallet.createCoinbase(
        coinbase_amount, path=alice_path)

    num_change_outputs = 1
    amount = 30000000000
    fee_base = 7000000
    block_height = 99999

    # prepare send transaction
    send_slate, alice_secret_key, alice_secret_nonce, change_outputs = alice_wallet.send(
        [output], num_change_outputs, amount, fee_base, block_height,
        path=alice_path, receiver_address=bob_address)
    send_slate_payload = send_slate.serialize()

    # prepare S1 send slatepack
    version = SlatepackVersion(0, 0)
    metadata = SlatepackMetadata(
        sender=SlatepackAddress.fromBech32(alice_address))
    emode = EMode.PLAINTEXT
    send_slatepack_message = SlatepackMessage(
        version, metadata, emode, send_slate_payload)
    assert not send_slatepack_message.is_encrypted()

    # encrypt the S1 slatepack
    send_slatepack_message.encryptPayload(
        [SlatepackAddress.fromBech32(bob_address)])
    slatepack_S1 = send_slatepack_message.pack()
    assert send_slatepack_message.is_encrypted()
    # now Alice sends slatepack_S1 to Bob

    # Bob receives it
    received_slatepack_S1 = bob_wallet.decryptSlatepack(
        slatepack_S1, path=bob_path)
    assert received_slatepack_S1.payload == send_slate_payload
    received_send_slate = received_slatepack_S1.getSlate()

    # now Bob has decrypted slate that Alice sent him
    # he is preparing his own slate
    receive_slate, bob_secret_key, bob_secret_nonce = bob_wallet.receive(received_send_slate)
    receive_slate_payload = receive_slate.serialize()

    # prepare slatepack_S2 for Alice
    version = SlatepackVersion(0, 0)
    metadata = SlatepackMetadata(
        sender=SlatepackAddress.fromBech32(bob_address))
    emode = EMode.PLAINTEXT
    receive_slatepack_message = SlatepackMessage(
        version, metadata, emode, receive_slate_payload)
    assert not receive_slatepack_message.is_encrypted()

    receive_slatepack_message.encryptPayload(
        [SlatepackAddress.fromBech32(alice_address)])
    slatepack_S2 = receive_slatepack_message.pack()
    # now Bob sends slatepack_S2 back to Alice

    # Alice receive it
    received_receive_slatepack_message = alice_wallet.decryptSlatepack(
        slatepack_S2, path=alice_path)
    assert received_receive_slatepack_message.payload == receive_slate_payload
    received_receive_slate = received_receive_slatepack_message.getSlate()

    # now Alice has decrypted slate that Bob sent her
    # she is ready to finalize the transaction
    finalized_slate = alice_wallet.finalize(
        received_receive_slate, path=alice_path)
    # this is the final slate that can be broadcasted to nodes

def test_srs_flow_slatepacks_persistent():
    mock_node = MockNode()

    # prepare wallets
    alice_path = 'm/0/1/0'
    alice_wallet = PersistentWallet(
        Wallet.initialize(), WalletStorageInMemory(), mock_node)
    alice_recovery_phrase = alice_wallet.getSeedPhrase()

    bob_path = 'm/0/1/0'
    bob_wallet = PersistentWallet(
        Wallet.initialize(), WalletStorageInMemory(), mock_node)
    bob_recovery_phrase = bob_wallet.getSeedPhrase()

    # their tip heights are at zero
    assert alice_wallet.get_tip_height() == 0
    assert bob_wallet.get_tip_height() == 0

    # while node is at 100000
    assert mock_node.get_current_block_height() == 100000

    # transaction setup
    fee_base = 7000000

    # check balance, every wallet is empty
    assert alice_wallet.balance().toJSON() == {
        'locked': 0,
        'awaiting_confirmation': 0,
        'spendable': 0,
        'total': 0
    }
    assert bob_wallet.balance().toJSON() == {
        'locked': 0,
        'awaiting_confirmation': 0,
        'spendable': 0,
        'total': 0
    }

    # Alice needs some coinbase to send to Bob
    coinbase_amount = 60000000000
    coinbase_transaction = alice_wallet.createCoinbase(
        coinbase_amount, path=alice_path)

    # submit the coinbase transaction to the node
    mock_node.push_transaction(coinbase_transaction)

    # prepare S1 send slatepack, but Alice has no confirmed funds yet
    with pytest.raises(Exception, match='Insufficient funds'):
        alice_wallet.send(
            30000000000,
            bob_wallet.getSlatepackAddress(),
            fee_base=fee_base)

    # mine a block to confirm Alice's coinbase
    mock_node.mine()

    # now Alice needs to refresh her wallet to detect confirmed funds
    alice_wallet.refresh()

    # Alice should have her confirmed coinbase balance now
    assert alice_wallet.balance().toJSON() == {
        'locked': 0,
        'awaiting_confirmation': 0,
        'spendable': 60000000000,
        'total': 60000000000
    }

    # now Alice can prepare S1 send slatepack
    s1_slatepack_message = alice_wallet.send(
        30000000000,
        bob_wallet.getSlatepackAddress(),
        fee_base=fee_base)
    s1_slatepack_text = s1_slatepack_message.pack() # this text gets sent to Bob

    # extract slate from S1 slatepack to check fee
    # we use Bob's wallet as Alice encrypted the slatepack for Bob
    slate_fee = bob_wallet.decryptSlatepack(s1_slatepack_message).getSlate().getFee().getFee()
    assert slate_fee == 322000000

    # check Alice's wallet outputs, the coinbase should be there but unconfirmed
    # also we expect main output to be locked
    assert alice_wallet.balance().toJSON() == {
        'locked': 60000000000,
        'awaiting_confirmation': 30000000000 - slate_fee,
        'spendable': 0,
        'total': 30000000000 - slate_fee
    }

    # now Bob receives S1 slatepack
    s1_slatepack_message_received = SlatepackMessage.unarmor(s1_slatepack_text)
    assert s1_slatepack_message_received.is_encrypted()
    s2_slatepack_message = bob_wallet.receive(s1_slatepack_message_received, path=bob_path)
    s2_slatepack_text = s2_slatepack_message.pack()  # this text gets sent back to Alice

    # check Bob's wallet outputs, the incoming funds should be there but unconfirmed
    assert bob_wallet.balance().toJSON() == {
        'locked': 0,
        'awaiting_confirmation': 30000000000,
        'spendable': 0,
        'total': 30000000000
    }

    # now Alice receives S2 slatepack
    s2_slatepack_message_received = SlatepackMessage.unarmor(s2_slatepack_text)
    assert s2_slatepack_message_received.is_encrypted()
    finalized_slate = alice_wallet.finalize(
        s2_slatepack_message_received, path=alice_path)

    # Alice sends transaction to node
    alice_wallet.push_finalized_slatepack(finalized_slate)

    # mine a block to confirm the transaction
    mock_node.mine()

    # they both check their wallets
    alice_wallet.refresh()
    bob_wallet.refresh()

    # check their balances
    alice_balance = alice_wallet.balance().toJSON()
    assert alice_balance == {
        'locked': 0,
        'awaiting_confirmation': 0,
        'spendable': 30000000000 - slate_fee,
        'total': 30000000000 - slate_fee
    }

    bob_balance = bob_wallet.balance().toJSON()
    assert bob_balance == {
        'locked': 0,
        'awaiting_confirmation': 0,
        'spendable': 30000000000,
        'total': 30000000000
    }

    # test restoring wallet from seed phrase and syncing with node
    # first Alice
    alice_wallet_restored = PersistentWallet.restoreFromSeedPhrase(
        alice_recovery_phrase, WalletStorageInMemory(), mock_node)
    assert not alice_wallet_restored.is_synced()

    alice_wallet_restored.scan()

    alice_balance_restored = alice_wallet_restored.balance().toJSON()
    assert alice_balance_restored == {
        'locked': 0,
        'awaiting_confirmation': 0,
        'spendable': 30000000000 - slate_fee,
        'total': 30000000000 - slate_fee
    }
    assert alice_wallet_restored.is_synced()

    # now Bob
    bob_wallet_restored = PersistentWallet.restoreFromSeedPhrase(
        bob_recovery_phrase, WalletStorageInMemory(), mock_node)
    assert not bob_wallet_restored.is_synced()

    bob_wallet_restored.scan()

    bob_balance_restored = bob_wallet.balance().toJSON()
    assert bob_balance_restored == {
        'locked': 0,
        'awaiting_confirmation': 0,
        'spendable': 30000000000,
        'total': 30000000000
    }

    assert bob_wallet_restored.is_synced()




import hmac
import os

from typing import Tuple, List, Optional, Dict, Union
from uuid import UUID

from nacl import bindings

from Crypto.Cipher import ChaCha20_Poly1305
from Crypto.Random import get_random_bytes

from hashlib import blake2b, pbkdf2_hmac, sha512

from bip32 import BIP32
from bip_utils import Bech32Encoder

from mimblewimble.crypto.rangeproof import RangeProof
from mimblewimble.models.slatepack.metadata import SlatepackVersion, SlatepackMetadata
from mimblewimble.serializer import Serializer
from mimblewimble.mnemonic import Mnemonic
from mimblewimble.keychain import KeyChain

from mimblewimble.crypto.aggsig import AggSig
from mimblewimble.crypto.commitment import Commitment
from mimblewimble.crypto.bulletproof import EBulletproofType
from mimblewimble.crypto.pedersen import Pedersen
from mimblewimble.crypto.secret_key import SecretKey

from mimblewimble.models.transaction import EOutputStatus
from mimblewimble.models.transaction import EOutputFeatures
from mimblewimble.models.transaction import EKernelFeatures
from mimblewimble.models.transaction import BlindingFactor
from mimblewimble.models.transaction import TransactionInput
from mimblewimble.models.transaction import TransactionOutput
from mimblewimble.models.transaction import TransactionBody
from mimblewimble.models.transaction import Transaction
from mimblewimble.models.transaction import TransactionKernel
from mimblewimble.models.fee import Fee

from mimblewimble.models.slatepack.address import SlatepackAddress
from mimblewimble.models.slatepack.message import SlatepackMessage, EMode

from mimblewimble.entity import OutputDataEntity
from mimblewimble.helpers.fee import calculateFee

from mimblewimble.slatebuilder import Slate, ESlateStage
from mimblewimble.slatebuilder import SendSlateBuilder
from mimblewimble.slatebuilder import ReceiveSlateBuilder
from mimblewimble.slatebuilder import FinalizeSlateBuilder
from mimblewimble.slatebuilder.invoice import InvoiceSlateBuilder
from mimblewimble.slatebuilder.pay import PaySlateBuilder


class Wallet:
    def __init__(self, encrypted_seed=None, salt=None, nonce=None, master_seed=None, tag=None):
        self.encrypted_seed = encrypted_seed
        self.salt = salt
        self.nonce = nonce
        self.master_seed = master_seed # 32 bytes


    def shieldWallet(self, passphrase: str, salt=None, nonce=None):
        if self.master_seed is None:
            raise Exception('The wallet is already shielded')
        if self.salt is None:
            if salt is None:
                self.salt = os.urandom(8)
            else:
                self.salt = salt
        if self.nonce is None:
            if nonce is None:
                self.nonce = os.urandom(12)
            else:
                self.nonce = nonce
        # compute encrypted_seed from master_seed
        key = pbkdf2_hmac('sha512', bytes(passphrase, 'utf-8'), self.salt, 100)
        cipher = ChaCha20_Poly1305.new(key=key[0:32], nonce=self.nonce)
        ciphertext = cipher.encrypt(self.master_seed)
        tag = cipher.digest()
        self.encrypted_seed = ciphertext + tag
        self.master_seed = None


    def unshieldWallet(self, passphrase: str, salt=None, nonce=None):
        if self.salt is None and salt is None:
            raise Exception('Missing salt')
        elif self.salt is None and salt is not None:
            self.salt = salt
        if self.nonce is None and nonce is None:
            raise Exception('Missing nonce')
        elif self.nonce is None and nonce is not None:
            self.nonce = nonce
        # compute master_seed from encrypted_seed
        key = pbkdf2_hmac('sha512', bytes(passphrase, 'utf-8'), salt, 100)
        cipher = ChaCha20_Poly1305.new(key=key[0:32], nonce=nonce)
        ciphertext = self.encrypted_seed[:32]
        tag = self.encrypted_seed[32:]
        plaintext = cipher.decrypt_and_verify(ciphertext, tag)
        self.master_seed = plaintext
        del plaintext


    def getSeedPhrase(self):
        if self.master_seed is None:
            raise Exception('The wallet is shielded')
        M = Mnemonic()
        return M.mnemonicFromEntropy(self.master_seed)


    def getEncryptedSeed(self):
        if None in [self.encrypted_seed, self.salt, self.nonce]:
            raise Exception('The wallet is not shielded')
        return {
            'encrypted_seed': self.encrypted_seed.hex(),
            'salt': self.salt.hex(),
            'nonce': self.nonce.hex()
        }


    def getSlatepackAddress(self, path='m/0/1/0', testnet=False):
        if self.master_seed is None:
            raise Exception('The wallet is shielded')

        keychain = KeyChain.fromSeed(self.master_seed)
        slatepack_address = keychain.deriveSlatepackAddress(path)

        return slatepack_address


    def createCoinbase(self, amount, path='m/0/1/0'):
        if self.master_seed is None:
            raise Exception('The wallet is shielded')

        # instantiate the keychain for this wallet seed
        keychain = KeyChain.fromSeed(self.master_seed)

        # we will be needing Pedersen commitments
        # and signature aggregation
        p = Pedersen()
        agg = AggSig()

        # blinding factor and Pedersend commitment
        blinding_factor = keychain.derivePrivateKeyAmount(
            path, amount)
        commitment = p.commit(amount, blinding_factor)

        # bulletproof
        rangeproof = keychain.generateRangeProof(
            path, amount, commitment,
            blinding_factor, EBulletproofType.ENHANCED)

        # add the commitments
        transparent_factor = BlindingFactor(bytes([0x00 for j in range(32)]))
        transparent_commitment = p.commit(amount, transparent_factor)
        kernel_commitment = p.commitSum([commitment], [transparent_commitment])

        # build the output
        transaction_output = TransactionOutput(
            EOutputFeatures.COINBASE_OUTPUT,
            commitment,
            rangeproof)

        # build the output data entity
        output_data_entity = OutputDataEntity(
            path, blinding_factor, transaction_output, amount,
            EOutputStatus.NO_CONFIRMATIONS)

        # build the coinbase signature
        message = blake2b(
            EKernelFeatures.COINBASE_KERNEL.to_bytes(1, 'big'),
            digest_size=32).digest()
        coinbase_signature = agg.buildSignature(
            blinding_factor, kernel_commitment, message)

        # build the kernel
        shift = 0
        fee = 0
        transaction_kernel = TransactionKernel(
            EKernelFeatures.COINBASE_KERNEL,
            Fee(shift, fee),
            0,
            kernel_commitment,
            coinbase_signature)

        # clean-up
        del p
        del agg

        # return the result
        return transaction_kernel, output_data_entity


    def createBlindedOutput(
            self,
            amount: int,
            bulletproof_type: EBulletproofType,
            path='m/0/1/0',
            wallet_tx_id=None):
        # Pedersen commitment class
        p = Pedersen()

        # instantiate the keychain for this wallet seed
        keychain = KeyChain.fromSeed(self.master_seed)

        blinding_factor = keychain.derivePrivateKeyAmount(
            path, amount)
        commitment = p.commit(amount, blinding_factor)

        # bulletproof
        rangeproof = keychain.generateRangeProof(
            path, amount, commitment,
            blinding_factor, bulletproof_type)

        # build the output
        transaction_output = TransactionOutput(
            EOutputFeatures.DEFAULT,
            commitment,
            rangeproof)

        # clean-up
        del p

        return OutputDataEntity(
            path, blinding_factor, transaction_output, amount,
            EOutputStatus.NO_CONFIRMATIONS, wallet_tx_id=wallet_tx_id)


    def send(
            self,
            inputs: List[OutputDataEntity],
            num_change_outputs: int,
            amount: int,
            fee_base: int,
            block_height: int,
            send_entire_balance=False,
            receiver_address=None,
            path='m/0/1/0',
            wallet_tx_id=None,
            testnet=False) -> Tuple[Slate, SecretKey, SecretKey, List[OutputDataEntity]]:
        # if entire balance is sent, no change outputs
        if send_entire_balance:
            change_outputs = 0

        # the total number of outputs and number of kernels
        num_total_outputs = 1 + num_change_outputs
        num_kernels = 1

        # calculate fee
        fee = calculateFee(fee_base, len(inputs), num_total_outputs, num_kernels)

        # compute the total input
        input_total = 0
        for input_entry in inputs:
            input_total += input_entry.getAmount()

        # adjust the send amount if entire balance is being sent
        if send_entire_balance:
            amount = input_total - fee

        # compute the change amount and prepare the change outputs
        # each with own blinding factor xC
        change_amount = input_total - (amount + fee)
        change_outputs = []
        for i in range(num_change_outputs):
            coin_amount = int(change_amount / num_change_outputs)
            if i == 0:
                coin_amount += change_amount % num_change_outputs
            change_outputs.append(self.createBlindedOutput(
                coin_amount, EBulletproofType.ENHANCED,
                path=path, wallet_tx_id=wallet_tx_id))

        # prepare sender and receiver address for the payment proof
        keychain = KeyChain.fromSeed(self.master_seed)
        sender_address_bytes = keychain.deriveED25519PublicKey(path)
        receiver_address_bytes = None
        if receiver_address is not None:
            receiver = SlatepackAddress.fromBech32(receiver_address)
            receiver_address_bytes = receiver.toED25519()

        # build send slate
        slate_builder = SendSlateBuilder(self.master_seed)
        slate, secret_key, secret_nonce = slate_builder.build(
            amount,
            fee,
            block_height,
            inputs,
            change_outputs,
            sender_address=sender_address_bytes,
            receiver_address=receiver_address_bytes,
            testnet=testnet)
        return slate, secret_key, secret_nonce, change_outputs


    def receive(
            self,
            send_slate: Slate,
            path='m/0/1/0',
            wallet_tx_id=None,
            testnet=False) -> Slate:
        output = self.createBlindedOutput(
            send_slate.getAmount(),
            EBulletproofType.ENHANCED,
            path=path,
            wallet_tx_id=wallet_tx_id)
        # build the receive slate
        slate_builder = ReceiveSlateBuilder(self.master_seed)
        receive_slate, secret_key, secret_nonce = slate_builder.addReceiverData(
            send_slate,
            output,
            testnet=testnet)

        # update the payment proof
        payment_proof = receive_slate.getPaymentProof()
        if payment_proof is not None:
            sender_address = payment_proof.getSenderAddress()
            # prepare the receiver address to be able to update
            # the payment proof
            keychain = KeyChain.fromSeed(self.master_seed)
            receiver_address = keychain.deriveED25519PublicKey(path)

            # prepare the kernel commitment
            kernel_commitment = receive_slate.getKernelCommitment().getBytes()

            # message to be signed:
            # (amount | kernel commitment | sender address)
            serializer = Serializer()
            serializer.write(receive_slate.getAmount().to_bytes(8, 'big'))
            serializer.write(kernel_commitment)
            serializer.write(sender_address)
            message = serializer.readall()

            # produce the signature
            receiver_signature = keychain.signED25519(message, path)

            # update the payment proof data
            receive_slate.proof_opt.setReceiverAddress(receiver_address)
            receive_slate.proof_opt.setReceiverSignature(receiver_signature)

        # done!
        return receive_slate, secret_key, secret_nonce


    def invoice(self,
            amount: int,
            wallet_tx_id=None,
            path: str = "m/0/1/0",
            block_height=0) -> Tuple[Slate, SecretKey, SecretKey]:
        output = self.createBlindedOutput(
            amount,
            EBulletproofType.ENHANCED,
            path=path,
            wallet_tx_id=wallet_tx_id)
        slate_builder = InvoiceSlateBuilder(self.master_seed)
        slate, secret_key, secret_nonce = slate_builder.build(
            amount,
            output,
            block_height=block_height,
            slate_version=0)

        return slate, secret_key, secret_nonce


    def pay(
            self,
            invoice_slate: Slate,
            inputs: List[OutputDataEntity],
            num_change_outputs: int,
            fee_base: int,
            path='m/0/1/0',
            wallet_tx_id=None,
            testnet=False
    ) -> Tuple[Slate, SecretKey, SecretKey, List[OutputDataEntity]]:

        # the total number of outputs and number of kernels
        num_total_outputs = 1 + num_change_outputs
        num_kernels = 1

        # calculate fee
        fee = calculateFee(fee_base, len(inputs), num_total_outputs, num_kernels)

        # compute the total input
        input_total = 0
        for input_entry in inputs:
            input_total += input_entry.getAmount()

        # extract amount
        amount = invoice_slate.getAmount()

        # compute the change amount and prepare the change outputs
        # each with own blinding factor xC
        change_amount = input_total - (amount + fee)
        change_outputs = []
        for i in range(num_change_outputs):
            coin_amount = int(change_amount / num_change_outputs)
            if i == 0:
                coin_amount += change_amount % num_change_outputs
            change_outputs.append(
                self.createBlindedOutput(
                    coin_amount,
                    EBulletproofType.ENHANCED,
                    path=path,
                    wallet_tx_id=wallet_tx_id
                )
            )

        # build the paid slate
        slate_builder = PaySlateBuilder(self.master_seed)
        slate, secret_key, secret_nonce = slate_builder.build(
            invoice_slate,
            fee,
            inputs,
            change_outputs)
        return slate, secret_key, secret_nonce, change_outputs

    def finalize(
            self,
            receive_slate: Slate,
            path='m/0/1/0',
            wallet_tx_id=None,
            testnet=False):
        # prepare the original sender address for the payment proof validation
        keychain = KeyChain.fromSeed(self.master_seed)

        check_payment_proof = receive_slate.stage == ESlateStage.STANDARD_RECEIVED

        # build the receive slate
        slate_builder = FinalizeSlateBuilder(self.master_seed)
        finalized_slate = slate_builder.finalize(
            receive_slate,
            testnet=testnet)

        if check_payment_proof:
            sender_address_bytes = keychain.deriveED25519PublicKey(path)
            slate_builder.verifyPaymentProof(
                finalized_slate,
                sender_address_bytes)

        return finalized_slate

    def ageDecrypt(self, ciphertext: bytes, path='m/0/1/0'):
        keychain = KeyChain.fromSeed(self.master_seed)
        return keychain.ageDecrypt(ciphertext, path)

    def deriveAgeSecretKey(self, path='m/0/1/0'):
        keychain = KeyChain.fromSeed(self.master_seed)
        return keychain.deriveAgeSecretKey(path)

    def decryptSlatepack(self, armored_slatepack_: Union[str, SlatepackMessage], path='m/0/1/0'):
        if isinstance(armored_slatepack_, str):
            slatepack_message = SlatepackMessage.unarmor(armored_slatepack_)
        else:
            slatepack_message = armored_slatepack_

        if not slatepack_message.is_encrypted():
            # TODO raise a warning it was not encrypted
            return slatepack_message

        keychain = KeyChain.fromSeed(self.master_seed)
        age_secret_key = keychain.deriveAgeSecretKey(path)

        slatepack_message.decryptPayload(age_secret_key)

        return slatepack_message


    @classmethod
    def initialize(self):
        master_seed = os.urandom(32)
        return Wallet(master_seed=master_seed)


    @classmethod
    def fromSeedPhrase(self, phrase: str):
        M = Mnemonic()
        master_seed = M.entropyFromMnemonic(phrase)
        return Wallet(master_seed=master_seed)


    @classmethod
    def fromEncryptedSeedDict(self, seed: dict, passphrase=None):
        return self.fromEncryptedSeed(
            seed['encrypted_seed'], seed['salt'], seed['nonce'],
            passphrase=passphrase)


    @classmethod
    def fromEncryptedSeed(
            self, encrypted_seed_hex: str, salt_hex: str, nonce_hex: str,
            passphrase=None):
        encrypted_seed = bytes.fromhex(encrypted_seed_hex)
        nonce = bytes.fromhex(nonce_hex)
        salt = bytes.fromhex(salt_hex)
        w = Wallet(encrypted_seed=encrypted_seed, salt=salt, nonce=nonce)
        if passphrase is not None:
            w.unshieldWallet(passphrase, nonce=nonce, salt=salt)
        return w

class WalletStorage:
    # Outputs
    def get_all_outputs(self) -> List[OutputDataEntity]:
        raise NotImplementedError

    def get_outputs_by_status(self, status: EOutputStatus) -> List[OutputDataEntity]:
        raise NotImplementedError

    def is_output_known(self, commitment: bytes) -> bool:
        raise NotImplementedError

    def add_output(self, output: OutputDataEntity, kernel: Optional[TransactionKernel] = None):
        """Add new output (e.g., receive or coinbase). Kernel optional for coinbase/confirmed."""
        raise NotImplementedError

    def update_output(self, output: OutputDataEntity):
        """Update status, height, etc. (e.g., lock, spend, confirm, mature)."""
        raise NotImplementedError

    def mark_output_spent(self, commitment: bytes):
        """Helper for refresh: mark as spent."""
        raise NotImplementedError

    # Transactions / Slates
    def save_slate_context(
        self,
        slate_id: UUID | str,
        slate,
        secret_key: Optional[bytes | SecretKey] = None,
        secret_nonce: Optional[bytes | SecretKey] = None,
        kernel: Optional[TransactionKernel] = None
    ):
        """Save in-progress or confirmed slate + blinding secrets."""
        raise NotImplementedError

    def get_slate_context(self, slate_id: UUID) -> Optional[Dict]:
        """Return dict with slate, secrets, kernel."""
        raise NotImplementedError

    def get_tx_kernel(self, slate_id: UUID) -> Optional[TransactionKernel]:
        """Get kernel for confirmed tx."""
        raise NotImplementedError

    def delete_slate_context(self, slate_id: UUID):
        """Cleanup after cancel or full confirm."""
        raise NotImplementedError

    # Misc
    def commit(self):
        """Flush changes (useful for DB impl)."""
        pass

class WalletStorageInMemory(WalletStorage):
    def __init__(self):
        self.outputs: List[OutputDataEntity] = []  # All outputs (active + spent)
        self.txs: Dict[UUID, Dict] = {}  # slate_id -> context

    # Outputs
    def get_all_outputs(self) -> List[OutputDataEntity]:
        return self.outputs[:]

    def get_outputs_by_status(self, status: EOutputStatus) -> List[OutputDataEntity]:
        return [o for o in self.outputs if o.status == status]

    def is_output_known(self, commitment: bytes) -> bool:
        for o in self.outputs:
            if o.output.commitment.getBytes() == commitment:
                return True
        return False

    def add_output(self, output: OutputDataEntity, kernel: Optional[TransactionKernel] = None):
        # For coinbase: set status=IMMATURE, link kernel if needed
        self.outputs.append(output)

        # TODO
        #if kernel and output.is_coinbase:  # assume flag or check features
        #    # Optional: store kernel separately or in output
        #    pass

    def update_output(self, updated_output: OutputDataEntity):
        for i, o in enumerate(self.outputs):
            if o.output.commitment == updated_output.output.commitment:  # assume unique commitment
                self.outputs[i] = updated_output
                return
        raise ValueError("Output not found")

    def mark_output_spent(self, commitment: bytes):
        for o in self.outputs:
            if o.output.commitment == commitment:
                o.output.status = EOutputStatus.SPENT
                return

    # Slates / Txs
    def save_slate_context(
        self,
        slate_id: UUID,
        slate,
        secret_key=None,
        secret_nonce=None,
        kernel=None
    ):
        self.txs[slate_id] = {
            'slate': slate,
            'secret_key': secret_key,
            'secret_nonce': secret_nonce,
            'kernel': kernel  # None for in-progress, set on finalize/post
        }

    def get_slate_context(self, slate_id: UUID) -> Optional[Dict]:
        return self.txs.get(slate_id)

    def get_tx_kernel(self, slate_id: UUID) -> Optional[TransactionKernel]:
        ctx = self.txs.get(slate_id)
        return ctx['kernel'] if ctx else None

    def delete_slate_context(self, slate_id: UUID):
        self.txs.pop(slate_id, None)

class NodeAccess:
    def get_current_block_height(self):
        raise NotImplementedError

    # methods consistent with RFC
    # https://docs.grin.mw/grin-rfcs/text/0007-node-api-v2/
    def get_outputs(self, outputs: List[OutputDataEntity]):
        raise NotImplementedError

    def get_unspent_outputs(self, start_index, end_index, window_size, range_proof=False):
        """Get unspent outputs in PMMR index range."""
        raise NotImplementedError

    def get_pmmr_indices(self, start_block_height: int, end_block_height: int) -> Dict[int, List[int]]:
        """Get MMR indices for outputs and kernels in block range."""
        raise NotImplementedError

    def get_kernel(self, excess: bytes) -> TransactionKernel:
        """Check if kernel exists, return kernel info + height if found."""
        raise NotImplementedError

    def push_transaction(self, transaction: Transaction):
        raise NotImplementedError

class WalletBalance:
    def __init__(self, locked, awaiting_confirmation, spendable):
        self.locked = locked
        self.awaiting_confirmation = awaiting_confirmation
        self.spendable = spendable
        self.total = awaiting_confirmation + spendable

    def toJSON(self):
        return {
            'locked': self.locked,
            'awaiting_confirmation': self.awaiting_confirmation,
            'spendable': self.spendable,
            'total': self.total
        }

class PersistentWallet:
    def __init__(
            self,
            wallet: Wallet,
            storage: WalletStorage,
            node=None,
            birthday_height=0):
        self._wallet = wallet
        self._storage = storage

        self.birthday_height = birthday_height
        self.tip_height = birthday_height

        if node is not None:
            self._node = node
        else:
            # TODO initialize the default local API node
            pass

    def get_tip_height(self):
        return self.tip_height

    def is_synced(self):
        return self.tip_height == self._node.get_current_block_height()

    def scan(self, window_size=100, sync_callback=None, path='m/0/1/0'):
        node_tip_height = self._node.get_current_block_height()
        if self.tip_height == node_tip_height:
            return

        # get the PMMR indices
        pmmr_indices_response = self._node.get_pmmr_indices(
            self.tip_height, node_tip_height)
        start_index = pmmr_indices_response['last_retrieved_index']
        end_index = pmmr_indices_response['highest_index']

        if sync_callback is not None:
            sync_callback.msg('pmmr_indices', {
                'start_index': start_index,
                'end_index': end_index
            })

        # scan the PMMR indices for outputs and kernels
        keychain = KeyChain.fromSeed(self._wallet.master_seed)
        while True:
            outputs_response = self._node.get_unspent_outputs(
                start_index, end_index, window_size, range_proof=True)
            unspent_outputs = outputs_response['outputs']

            for unspent_output in unspent_outputs:
                commitment_bytes = bytes.fromhex(unspent_output['commit'])
                commitment = Commitment(commitment_bytes[1:]) # remove the "09" prefix

                proof_bytes = bytes.fromhex(unspent_output['proof'])
                range_proof = RangeProof.frombytes(proof_bytes)

                try:
                    # TODO should path be taken into account?
                    rewind_proof = keychain.rewindRangeProof(
                        commitment,
                        range_proof,
                        EBulletproofType.ENHANCED)
                except Exception as e:
                    # this is not our output, skip
                    continue

                if sync_callback is not None:
                    sync_callback.msg('new_output', {
                        'commitment': commitment_bytes.hex()
                    })

                # this output belongs to this wallet
                # we should add it to the storage
                transaction_output = TransactionOutput(
                    EOutputFeatures.DEFAULT,
                    commitment,
                    range_proof)

                # build the output data entity
                output_data_entity = OutputDataEntity(
                    path,
                    rewind_proof.blinding_factor,
                    transaction_output,
                    rewind_proof.amount,
                    EOutputStatus.SPENDABLE)

                # save this output as spendable in the local storage
                if not self._storage.is_output_known(commitment.getBytes()):
                    self._storage.add_output(output_data_entity)
                else:
                    self._storage.update_output(output_data_entity)

            if outputs_response['highest_index'] <= outputs_response['last_retrieved_index']:
                break
            start_index = outputs_response['last_retrieved_index'] + 1

        self.tip_height = node_tip_height

    def refresh(self):
        # get all outputs awaiting confirmation
        # outputs = self._storage.get_outputs_by_status(EOutputStatus.NO_CONFIRMATIONS)
        outputs = self._storage.get_all_outputs() # TODO make it more efficient
        if len(outputs) == 0:
            return

        # query node for their status
        commitments = []
        for output in outputs:
            if isinstance(output, OutputDataEntity):
                commit = output.output.getCommitment().getBytes()
            if isinstance(output, TransactionOutput):
                commit = output.getCommitment().getBytes()
            commitments.append(commit)
        result = self._node.get_outputs(commitments)

        # update outputs in storage
        for commitment_bytes, output_info in result.items():
            for output in outputs:
                if output.output.getCommitment().getBytes() == commitment_bytes:
                    # update output status based on node info
                    if output_info['spent']:
                        output.status = EOutputStatus.SPENT
                    else:
                        output.status = EOutputStatus.SPENDABLE
                    output.block_height = output_info['height']
                    output.mmr_index = output_info['mmr_index']
                    break

    def post_transaction(self, finalized_slate: SlatepackMessage):
        self._node.post_transaction(finalized_slate)

    def shieldWallet(self, passphrase: str, salt=None, nonce=None):
        self._wallet.shieldWallet(passphrase, salt=salt, nonce=nonce)

    def unshieldWallet(self, passphrase: str, salt=None, nonce=None):
        self._wallet.unshieldWallet(passphrase, salt=salt, nonce=nonce)

    def createCoinbase(self, amount, path='m/0/1/0'):
        kernel, output = self._wallet.createCoinbase(amount, path=path)

        offset = BlindingFactor(bytes([0x00 for j in range(32)]))

        inputs = [
            TransactionInput(
                EKernelFeatures.COINBASE_KERNEL,
                output.getCommitment()
            )
        ]
        outputs = [output]
        kernels = [kernel]
        transaction_body = TransactionBody(
            inputs, outputs, kernels
        )

        coinbase_transaction = Transaction(
            offset,
            transaction_body
        )

        # TODO add to storage transactions + output
        self._storage.add_output(output, kernel=kernel)

        return coinbase_transaction

    def identify_own_outputs_from_slate(
            self, slate, path='m/0/1/0', new_status=EOutputStatus.NO_CONFIRMATIONS):
        slate_context = self._storage.get_slate_context(slate.slate_id)
        kernel = slate_context['kernel']

        keychain = KeyChain.fromSeed(self._wallet.master_seed)
        for commitment in slate.commitments:
            if commitment.range_proof:
                # now we know this commitment is an output
                range_proof = commitment.range_proof
                # we will check if this output belongs to this wallet
                # this way we will know if we should store it
                try:
                    # TODO should path be taken into account?
                    rewind_proof = keychain.rewindRangeProof(
                        commitment.commitment,
                        range_proof,
                        EBulletproofType.ENHANCED) # ORIGINAL
                except Exception as e:
                    # this is not our output, skip
                    continue

                # this output belongs to this wallet
                # we should add it to the storage
                transaction_output = TransactionOutput(
                    EOutputFeatures.DEFAULT,
                    commitment.commitment,
                    range_proof)

                # build the output data entity
                output_data_entity = OutputDataEntity(
                    path,
                    rewind_proof.blinding_factor,
                    transaction_output,
                    rewind_proof.amount,
                    new_status)

                if not self._storage.is_output_known(commitment.commitment.getBytes()):
                    self._storage.add_output(output_data_entity, kernel=kernel)
                else:
                    self._storage.update_output(output_data_entity)

    def send(
            self,
            amount: int,
            receiver: SlatepackAddress,
            path='m/0/1/0',
            num_change_outputs=1,
            fee_base=0,
            block_height=None,
            encrypt=True):
        if block_height is None:
            block_height = self._node.get_current_block_height()

        spendable_outputs = self._storage.get_outputs_by_status(EOutputStatus.SPENDABLE)
        inputs = []
        total_amount = 0
        for output in spendable_outputs:
            if total_amount >= amount + fee_base:
                break
            inputs.append(output)
            total_amount += output.getAmount()
        if total_amount < amount + fee_base:
            raise Exception('Insufficient funds')

        send_slate, secret_key, secret_nonce, change_outputs = self._wallet.send(
            inputs,
            num_change_outputs,
            amount,
            fee_base,
            block_height,
            path=path,
            receiver_address=receiver)

        self._storage.save_slate_context(
            send_slate.slate_id,
            send_slate,
            secret_key=secret_key,
            secret_nonce=secret_nonce)

        # mark inputs as being locked/spent
        for input in inputs:
            input.status = EOutputStatus.LOCKED
            self._storage.update_output(input)

        # mark change outputs as NO_CONFIRMATIONS
        for output in change_outputs:
            output.status = EOutputStatus.NO_CONFIRMATIONS
            self._storage.add_output(output)

        # prepare s1 slatepack
        sender_address = self.getSlatepackAddress(path=path)
        version = SlatepackVersion(0, 0)
        metadata = SlatepackMetadata(
            sender=SlatepackAddress.fromBech32(sender_address))
        emode = EMode.PLAINTEXT
        send_slatepack_message = SlatepackMessage(
            version, metadata, emode, send_slate.serialize())

        if encrypt:
            send_slatepack_message.encryptPayload(
                [SlatepackAddress.fromBech32(receiver)])
            return send_slatepack_message

        return send_slatepack_message

    def receive(
            self,
            send_slate: SlatepackMessage,
            path='m/0/1/0',
            encrypt=True):
        if send_slate.is_encrypted():
            s1 = self._wallet.decryptSlatepack(
                send_slate, path=path)
        else:
            s1 = send_slate

        slate = s1.getSlate()
        receive_slate, secret_key, secret_nonce = self._wallet.receive(
            slate,
            path=path)

        self._storage.save_slate_context(
            receive_slate.slate_id,
            receive_slate,
            secret_key=secret_key,
            secret_nonce=secret_nonce
        )
        self.identify_own_outputs_from_slate(receive_slate, path=path)

        # prepare s2 slatepack
        receive_slate_payload = receive_slate.serialize()
        sender_address = self.getSlatepackAddress(path=path)
        version = SlatepackVersion(0, 0)
        metadata = SlatepackMetadata(
            sender=SlatepackAddress.fromBech32(sender_address))
        emode = EMode.PLAINTEXT
        receive_slatepack_message = SlatepackMessage(
            version, metadata, emode, receive_slate_payload)

        if encrypt:
            if not send_slate.metadata.sender:
                raise Exception('Cannot encrypt receive slatepack: original sender address missing')
            original_sender_address = send_slate.metadata.sender  # original sender address
            receive_slatepack_message.encryptPayload(
                [original_sender_address])

        return receive_slatepack_message

    def invoice(
            self,
            amount: int,
            receiver: SlatepackAddress,
            block_height=None,
            path='m/0/1/0', encrypt=True):
        if block_height is None:
            block_height = self._node.get_current_block_height()

        invoice_slate, secret_key, secret_nonce = self._wallet.invoice(
            amount,
            path=path,
            block_height=block_height)

        self._storage.save_slate_context(
            invoice_slate.slate_id,
            invoice_slate,
            secret_key=secret_key,
            secret_nonce=secret_nonce)
        self.identify_own_outputs_from_slate(invoice_slate, path=path)

        # prepare r1 slatepack
        sender_address = self.getSlatepackAddress(path=path)
        version = SlatepackVersion(0, 0)
        metadata = SlatepackMetadata(
            sender=SlatepackAddress.fromBech32(sender_address))
        emode = EMode.PLAINTEXT
        invoice_slatepack_message = SlatepackMessage(
            version, metadata, emode, invoice_slate.serialize())

        if encrypt:
            invoice_slatepack_message.encryptPayload(
                [SlatepackAddress.fromBech32(receiver)])
            return invoice_slatepack_message

        return invoice_slatepack_message

    def pay(
            self,
            invoice_slate: SlatepackMessage,
            path='m/0/1/0',
            num_change_outputs=1,
            fee_base=0,
            encrypt=True
    ):
        if invoice_slate.is_encrypted():
            r1 = self._wallet.decryptSlatepack(
                invoice_slate, path=path)
        else:
            r1 = invoice_slate

        slate = r1.getSlate()
        requested_amount = slate.getAmount()

        spendable_outputs = self._storage.get_outputs_by_status(EOutputStatus.SPENDABLE)
        inputs = []
        total_amount = 0
        for output in spendable_outputs:
            if total_amount >= requested_amount + fee_base:
                break
            inputs.append(output)
            total_amount += output.getAmount()
        if total_amount < requested_amount + fee_base:
            raise Exception('Insufficient funds')

        pay_slate, secret_key, secret_nonce, change_outputs = self._wallet.pay(
            slate,
            inputs,
            num_change_outputs,
            fee_base,
            path=path)

        self._storage.save_slate_context(
            pay_slate.slate_id,
            pay_slate,
            secret_key=secret_key,
            secret_nonce=secret_nonce)

        # mark inputs as being locked/spent
        for input in inputs:
            input.status = EOutputStatus.LOCKED
            self._storage.update_output(input)

        # mark change outputs as NO_CONFIRMATIONS
        for output in change_outputs:
            output.status = EOutputStatus.NO_CONFIRMATIONS
            self._storage.add_output(output)
        # self.identify_own_outputs_from_slate(pay_slate, path=path)

        # prepare r2 slatepack
        receive_slate_payload = pay_slate.serialize()
        sender_address = self.getSlatepackAddress(path=path)
        version = SlatepackVersion(0, 0)
        metadata = SlatepackMetadata(
            sender=SlatepackAddress.fromBech32(sender_address))
        emode = EMode.PLAINTEXT
        pay_slatepack_message = SlatepackMessage(
            version, metadata, emode, receive_slate_payload)

        if encrypt:
            if not invoice_slate.metadata.sender:
                raise Exception('Cannot encrypt receive slatepack: original sender address missing')
            original_sender_address = invoice_slate.metadata.sender  # original sender address
            pay_slatepack_message.encryptPayload(
                [original_sender_address])

        return pay_slatepack_message

    def finalize(
            self,
            final_slate: SlatepackMessage,
            path='m/0/1/0',
            testnet=False):
        if final_slate.is_encrypted():
            s2 = self._wallet.decryptSlatepack(
                final_slate, path=path)
        else:
            s2 = final_slate

        slate = s2.getSlate()
        slate_context = self._storage.get_slate_context(slate.slate_id)
        secret_key = slate_context['secret_key']
        secret_nonce = slate_context['secret_nonce']
        kernel = slate_context['kernel']
        finalized_slate = self._wallet.finalize(
            slate,
            path=path,
            testnet=testnet)

        self._storage.save_slate_context(
            finalized_slate.slate_id,
            finalized_slate,
            kernel=kernel,
            secret_key=secret_key,
            secret_nonce=secret_nonce
        )
        self.identify_own_outputs_from_slate(finalized_slate, path=path)

        return finalized_slate

    def push_finalized_slatepack(self, finalized_slate: Slate):
        inputs = []
        outputs = []
        kernels = []
        for commitment in finalized_slate.commitments:
            if commitment.range_proof:
                # this is output
                output = TransactionOutput(
                    EOutputFeatures.DEFAULT,
                    commitment.commitment,
                    commitment.range_proof)
                outputs.append(output)
            else:
                input = TransactionInput(
                    EOutputFeatures.DEFAULT,
                    commitment.commitment)
                inputs.append(input)

        kernel = TransactionKernel(
            EKernelFeatures.DEFAULT_KERNEL,
            Fee(0, finalized_slate.fee),
            finalized_slate.lock_height,
            finalized_slate.getKernelCommitment(),
            finalized_slate.getSignature(0) # 0 ? TODO
        )
        kernels.append(kernel)

        transaction_body = TransactionBody(
            inputs, outputs, kernels
        )

        transaction = Transaction(
            finalized_slate.offset,
            transaction_body
        )

        self._node.push_transaction(transaction)

    def balance(self):
        locked = 0
        awaiting_confirmation = 0
        spendable = 0

        outputs = self._storage.get_all_outputs()
        for output in outputs:
            if output.status == EOutputStatus.NO_CONFIRMATIONS:
                awaiting_confirmation += output.getAmount()
            elif output.status == EOutputStatus.LOCKED:
                locked += output.getAmount()
            elif output.status == EOutputStatus.SPENDABLE:
                spendable += output.getAmount()

        return WalletBalance(
            locked,
            awaiting_confirmation,
            spendable
        )

    def decryptSlatepack(self, armored_slatepack_: Union[str, SlatepackMessage], path='m/0/1/0'):
        return self._wallet.decryptSlatepack(armored_slatepack_, path=path)

    def ageDecrypt(self, ciphertext: bytes, path='m/0/1/0'):
        return self._wallet.ageDecrypt(ciphertext, path=path)

    def ageEncrypt(self, plaintext: bytes, receiver_address: str):
        keychain = KeyChain.fromSeed(self._wallet.master_seed)
        age_public_key = keychain.deriveAgePublicKeyFromSlatepackAddress(
            receiver_address)
        return keychain.ageEncrypt(plaintext, age_public_key)

    def getSlatepackAddress(self, path='m/0/1/0', testnet=False):
        return self._wallet.getSlatepackAddress(path=path, testnet=testnet)

    def getSeedPhrase(self):
        return self._wallet.getSeedPhrase()

    @classmethod
    def restoreFromSeedPhrase(
            self,
            phrase: str,
            storage: WalletStorage,
            node=None,
            birthday_height=0):
        return PersistentWallet(
            Wallet.fromSeedPhrase(phrase),
            storage,
            node=node,
            birthday_height=birthday_height)


import pytest

from mimblewimble.pow.algos.cuckaroo import cuckaroo_validate
from mimblewimble.pow.algos.cuckarood import cuckarood_validate
from mimblewimble.pow.algos.cuckaroom import cuckaroom_validate
from mimblewimble.pow.algos.cuckarooz import cuckarooz_validate
from mimblewimble.pow.algos.cuckatoo import cuckatoo_validate
from mimblewimble.pow.prepow import serialize_pre_pow
from mimblewimble.pow.common import EPoWStatus
from mimblewimble.pow.algos import pow_validate

from tests.fixtures import block_42_header

# cuckaroo
from tests.fixtures import block_174001_header as cuckaroo_header

# cuckarood
from tests.fixtures import block_275001_header as cuckarood_header

# cuckaroom
from tests.fixtures import block_600000_header as cuckaroom_header

# cuckarooz
from tests.fixtures import block_800015_header as cuckarooz_header

# cuckatoo
from tests.fixtures import block_604850_header as cuckatoo_header

from hashlib import blake2b


def test_cuckaroo_verification():
    edge_bits = cuckaroo_header['edge_bits']
    solution = cuckaroo_header['cuckoo_solution']
    pre_pow = serialize_pre_pow(cuckaroo_header)
    keybuf = blake2b(pre_pow, digest_size=32).digest()
    valid, msg = cuckaroo_validate(
        solution, keybuf, edge_bits)
    assert valid
    assert msg == EPoWStatus.POW_OK

def test_cuckarood_verification():
    solution = cuckarood_header['cuckoo_solution']
    pre_pow = serialize_pre_pow(cuckarood_header)
    keybuf = blake2b(pre_pow, digest_size=32).digest()
    valid, msg = cuckarood_validate(solution, keybuf)
    assert valid
    assert msg == EPoWStatus.POW_OK

def test_cuckaroom_verification():
    solution = cuckaroom_header['cuckoo_solution']
    pre_pow = serialize_pre_pow(cuckaroom_header)
    keybuf = blake2b(pre_pow, digest_size=32).digest()
    valid, msg = cuckaroom_validate(solution, keybuf)
    assert valid
    assert msg == EPoWStatus.POW_OK

def test_cuckarooz_verification():
    solution = cuckarooz_header['cuckoo_solution']
    pre_pow = serialize_pre_pow(cuckarooz_header)
    keybuf = blake2b(pre_pow, digest_size=32).digest()
    valid, msg = cuckarooz_validate(solution, keybuf)
    assert valid
    assert msg == EPoWStatus.POW_OK

def test_cuckatoo_verification():
    edge_bits = cuckatoo_header['edge_bits']
    solution = cuckatoo_header['cuckoo_solution']
    pre_pow = serialize_pre_pow(cuckatoo_header)
    keybuf = blake2b(pre_pow, digest_size=32).digest()
    valid, msg = cuckatoo_validate(solution, keybuf, edge_bits)
    assert valid
    assert msg == EPoWStatus.POW_OK

def test_pow_algo_auto():
    assert pow_validate(cuckaroo_header) == (True, EPoWStatus.POW_OK)
    assert pow_validate(cuckarood_header) == (True, EPoWStatus.POW_OK)
    assert pow_validate(cuckaroom_header) == (True, EPoWStatus.POW_OK)
    assert pow_validate(cuckarooz_header) == (True, EPoWStatus.POW_OK)
    assert pow_validate(cuckatoo_header) == (True, EPoWStatus.POW_OK)



from mimblewimble.models.slatepack.message import slatepack_pack
from mimblewimble.models.slatepack.message import slatepack_unpack
from mimblewimble.models.slatepack.message import slatepack_spacing

# example from the grin wiki
# https://docs.grin.mw/wiki/transactions/slatepack/
example = '''
BEGINSLATEPACK. 4H1qx1wHe668tFW yC2gfL8PPd8kSgv
pcXQhyRkHbyKHZg GN75o7uWoT3dkib R2tj1fFGN2FoRLY
GWmtgsneoXf7N4D uVWuyZSamPhfF1u AHRaYWvhF7jQvKx
wNJAc7qmVm9JVcm NJLEw4k5BU7jY6S eb.
ENDSLATEPACK.
'''[1:]

example_raw = bytes.fromhex(''.join([
    '4d047e46000400034bea',
    '91b819b84f0081f925b6',
    'edd81537010600000000',
    '4fb1004000000000007a',
    '120001000225418cd179',
    '18870cc7be5b629cd7e5',
    '3188e08dfb6b043cf112',
    '14af1e77b2ee730391b7',
    '9272fca73ed341e0c033',
    '935799b4493d3273064a',
    '4c2d6fa9c1c4050271cc',
    '00'
]))

# my own example
raw = bytes.fromhex(''.join([
    '0f28c78a7c2e6b9c1fe3',
    '7abfbf1baf8a161cb2cd',
    '95670e3091aaef9c9d5f',
    'dc2fa11338c352284cf6',
    '157fa05b2da153b581ab',
    '5076605d783add8d9e2d',
    '0dfb07d323f85f50a121',
    '05fe69e78edc338fcfa1',
    'd45e99054b31ac385a57',
    '94a934ea0bc977b7f949',
    '49180ec8f1dc10b6b177',
    '720ced24621dae7e5ab4',
    '5ef7413761cb4211'
]))

encoded_slatepack = '''
BEGINSLATEPACK. 2YyRpvomiRisAo6 uhtewqXgRsk3Cdk
rBduskjuuXSur5V gJTb911QXf9WdMe P5Rz4J3r15U6tmb
jFdV36xjimQDzmq M3C3DrRGUyNJu7x nDJTv8swDVnEgwN
QS4bcbA9DPmoJCE S6zfBNWkDoKzu9P z1ZaReA84i6173X
k6DZ6kiLhW. ENDSLATEPACK.
'''[1:]


def test_slatepack_armor_official():
    decoded = slatepack_unpack(example)
    assert decoded == example_raw

    packed = slatepack_pack(decoded, words_per_line=3)
    assert packed == example


def test_slatepack_armor():
    packed = slatepack_pack(raw, words_per_line=3)
    assert packed == encoded_slatepack

    decoded = slatepack_unpack(packed)
    assert decoded == raw




from mimblewimble.slatebuilder.slate import SettableByte


def test_settable_bytes():
    #        76543210
    # 0xb5 = 10110101 = 181
    b = SettableByte(0xb5)
    assert b.get_bit(0) == 1
    assert b.get_bit(1) == 0
    assert b.get_bit(2) == 1
    assert b.get_bit(3) == 0
    assert b.get_bit(4) == 1
    assert b.get_bit(5) == 1
    assert b.get_bit(6) == 0
    assert b.get_bit(7) == 1
    assert b.get_value() == 181

    #        76543210
    # 0xff = 11111111 = 255
    b.set_bit(1, 1)
    b.set_bit(3, 1)
    b.set_bit(6, 1)
    assert b.get_value() == 255

    #        76543210
    # 0x4a = 01001010 = 74
    b.set_bit(0, 0)
    b.set_bit(2, 0)
    b.set_bit(4, 0)
    b.set_bit(5, 0)
    b.set_bit(7, 0)
    assert b.get_value() == 74



from mimblewimble.models.slatepack.message import SlatepackMessage, EMode
from mimblewimble.wallet import Wallet

address_a = 'grin14kgku7l5x6te3arast3p59zk4rteznq2ug6kmmypf2d6z8md76eqg3su35'
address_b = 'grin1m4krnajw792zxfyldu79jssh0d3kzjtwpdn2wy7fysrfw4ej0waskurq76'

seed_a = 'sign interest obtain raw window monster jump bring nice crunch toward grunt prosper recycle sphere battle mother fold reject velvet emotion similar romance govern'
seed_b = 'rough spy shallow parade boil sausage toss year used senior infant hen impulse prefer divorce flame drink save tackle front record scrub march doctor'

ba_s1 = '''
BEGINSLATEPACK. auBNnRrahGnJ1iw 52RSppvCNzAPyT8 p5icMiMDYjDKbHH
8gd9Xci3AWGMd88 PWt36uc7uPVKocB SnxB28ptvgmfEn3 SouRUUBnjSEQCGi
gkwuzswEKLict2X A7sc7Rdu21gMFec Eq5AmyExTCjPHYg CU1DWQZC28kab8y
Fu1meQA5sYUQWM7 rvg1yADen6Z8R4S b3eVPg54eYwNv17 XqV1Lc3ACLSHycK
Gc7dPmAmBeZ7RxY JLdteR1QtFu8ngu GHSTNrui3TVkKug QJuN34WsJcCZWFc
AYKSYdBnwdXSPYy LsPCS3n4Mqo52HP U8kCq7sHsBdBbjV 9dcFQrm18pvWxVR
GJNm8XSrQtK9dyQ JvZxjv7UNTvh8q1 5yDXLA7z8L6NV2m dHZ6ujtecsSZdF5
mZqZyjsxeoj9kDr jjAXPD6gTVjobkh sjxXb1YU4qEfHnR wx7NjBx5RamzgEa
uWvARddnJd8pG2m wsptfQkvfBKogS5 1vRvmFMUb8MwPjW hucAnKcMaFLj1Hg
ESbV3HycopcL2VJ HJgmQQeFsqbyGMm Xkiz4sH2X6hWj1A D6rkR7uhDLL5YbY
MPwkFsRNK8zcPk7 X2DMCFZd5VNGcMZ gPBidMhw4nbUzii bj42vtLpT68JpTM
qeLUsuCzBUPs7h5 Z8vH9humdjkxkPM JK2z91cLcyWpqfN PCK2C96aWRmw1zK
vhA2bdauCkTRDD4 dES9RdcExqtMjby p2wvUFk2V2UHEw7 Sjpcc37kJ2P2ZG2
WYyM3VSXuMPSdwZ HVyvnXs4tSJTzsE 2wJTS5gRJX74FXW izjtA8tUGWcmCif
kBYie74pEWBAe4t ja5SLKkX4Ut1Ys8 rwPyB5zf6sGzrb7 3VyDtX85AmF7moE
MbaYHdP1tbM. ENDSLATEPACK.
'''

ba_s2 = '''
BEGINSLATEPACK. 8eeCbUNBQbXjjEa 9FMu2nM18wKWmgn Ychgtt7UqXkMnUk
7Y7NyvUqKqjXWj9 4xYUi1RwrqqKhqU c3xGHjGFDuFATDG LiqhJr3XJ453Qar
TrLgsU3juNeQto8 UfqHeNcjBM3obcu uUqnJH7UH1eRoV7 ehWLPQk5DQYWmMe
vsTqKABAo37U5Di uDUz6W7HCJvfWnW DJvHCTxqCPaTRwk WJMZx2AH9qdyaHw
PRtpoC11F7Cjxa8 HWb63pXRC5ieJVA LkZhXrA8jFb3S5k uuNvZKjetBLXJS2
8JDhAShCmQmM5kC Vyq3cAHi79yx5xq mAy3wRB4H7EogUU G1ih3MZx48YgrzA
sVcA68KgC5TLcLX 9EncMNbx7WbAwte prkNQZaL7JRCzt9 v1qsRBphS43d3kE
vuYsazrj15uSyuP sbn5LfjEEdXtTop AZ9ioi4XAkSohLx ytFbYpC7KSEwZsu
ijE7GY35RwVmhGa mAHUafWCy7Nk3NG GG2gxxG1uqejaYn Tx7qAm8VyFzQAsc
iWcALGmFu9g33Lu 2QkRfxhYjfqhrUA jhEzxrB2bzArsHG xpEh8S6ZrfS5e36
mZ6k6ZP71nwjre4 q18kGby8uw8o7bN Gc3chEYgyzBE3DE BvGWabEtE2Zpj8s
gG2Wj3Tg36YZBhv mP6CMGswgo2ZGSw ykBTeMbw57muTED wSUjeoJNfkK1thc
w6wCjwL4F6tzoWC NCHRc5esngfnoDs vLt6PMY3qHds3vu hYqte8jRsf2biHG
LrgFDbZ39NKYNeg LhrJctCHpee4Rye 3rGQK17KC7m3atV 7S926vmWp4LQVfZ
R8KbeZzEXkqjP4a FgntJUqVX7CJeAG r3cnrCyLsum4zEA gjRWiDtLZwXSypo
LuWrAu9oAYvHzK5 KQcK9H6GLho2Eus cJp3tqduRTmx97r 16PT6ZJ2PSQVyFW
egs3t4DEb4sKc6K bKBdw3MvnDeJTye yzgJtBiDpsGE4Tp F6NJbemA5bJfoPG
VNFeRYwm1hdHuTF p3chUTpMzr7EHZf q19MNQx1rhdzjnP wW2sN1ShsRHX9qu
owfwxmd8dTpSAXn hM6xujAWMonTNKT 1YL5AL28xeAJYpT a2uqFKaVx5y9D8M
cDK5pcY7wX8r2jc 9JXXoUKQ98JZLSW 7fGeFCAe7MrtPR1 t64FPwgYHgFkpd7
WrmJkfwWFShmcz4 1cb6aw93PZncen7 BLNyhAb9D99cd5g JjKGKyDnfMNjP4F
J7UWPR95cUsBoyX YToqZVn4evsG5vk sBsEaibznofuo6A WYGRVpNszPggEYH
UuEkbKHkyZdwoiP HdPyEVEz7mcrRQF sb1RTbiBYLy7syz hMCoaDBFBkHMGVw
vUFkfzdZ866USCP CQtRxKdmTbY6sSJ xwqnsRpKTXEmtSx 7oysfbY6zC9Rnfr
8WS1bUsqjaFZj4j fCFR2UD3iYDP5Kq 3xfeYLosZaAtouN 17pPv4X39zPpGGc
QeYHpgPEB581aJy CYRvmBdkuFuK4pY rLrnwpdNa5ugijM BLcozDYrHEV2WHU
nNfv6fAvjkB2Z3m dsji1im3mL2yJdy zcp4BVcG7WN3XDt m8D1GUSEocfaxTm
9JmmXGhAJuARz21 fqYpxs6HngxZAqd oiEKmFYqcACewnG TT55WR8wFsSV4wf
cwAYPmPX99G4SS1 heohousAXMwLxmG F4JBAKbsgKJUQTu jqiZfkhEzLCcZH2
fbFEjdVmZkVVMMq GzPg6Xt2ttXYBFh PwfkwCmH5tBvjXg pJDpVhybnbzXxSZ
E2itVArrj6MSYyN svFEmac817qZfjH NyVeq8DCuo2ScPw bBY7JhXh6j4pzeZ
sLGhvaFtVaARrsz 71eprF7HV9AjhLH jhpRruGeAgMxtpB C12JKLSXXqiBABT
UqRdwmMQTnDJKCz 7w5utNFyFbc335z kQzR8jvdPFSP1i. ENDSLATEPACK.
'''

ba_s3 = '''
BEGINSLATEPACK. 4a2nUgAKEK8d2Az ZdgMA1nwgiSn2th MpyxW1z1nwqZpma
ZBekrEDmSdz8e7s ekpGSBv9qtnL4vN yhLqpcwcSFuJoNt 2vdFzB3RCmas4qJ
UGnZ25D8qtZXEtN MDSUCv86cVkbeTv DMfU32mCEGNriP3 jmAhdUZZcXTMxXE
EXDUC3xpHUX4Fd2 FgvvGrwCPAoXzDr Yt5rzqhp4F4MmiE CJcPMAaWdx5LeqN
i1YBdQVf5Pr645G nojmue85VZrKtEs zR3guVVV2huqxXZ PUAdN6KLNswwaYC
VoTLVvEvmyrHQAu S33w5NG7wVYndpk Sn1u4XTZtftaryz QY2LWRNGvShz2we
RWKPWugDmtDftuj rvKbF8qAwWcRfu9 EoZnAiJQoEj1bJj TomGvdKCnEngzup
fmx4h5z7vcSSUVH nTN3S3FDU1Tosoc QjCHNp6Zn6wMRtj EWZT72CHLNP9aTK
EV6K6WGVo8QDPoJ FtifVaiu52T9T83 tzoR6VRVnPSdrXH EyTMBxxKQSSHbvH
SwDu4M37rC7ic5H 2i89sjnGunHJzHt wvAQQm7FToS53Fi UeMsnccPHeedwbs
h6TPnbznT5ZzyUE zH4fxH94mveCGHh 9X5W8DHNpWcqohS kw3yJBnDkvvuoec
QSstz1uWUGfA1Zh mq4cRnBJAjXQaqP 6aDL3XxmYo4oFrL 4rxeQZokPMwZKvc
3dZp9DqMGiQtZVH Qqqeik1M9SMjRnZ 924FMfrjgHTKabd Nbp1DMDHp9fbhmk
NykX81th5ntK9Jv 4P4XyPcbUUFWiFL hcBN2woqtH63Wwa m3PndM2oTTtpCZk
NZE1EymDXhdNBTW 7CyR8gZF52Cjw1K Tk4BNGefTJmzohd GyVg12RuqJUU9aS
cTyUEygg92HCA73 pKXas4rvYXeXh9v pdYWSLghjoSTUap 49Uh72TkDViU4bq
GqnYYMwFVEW5SFN xjiXk3JYgB9cqC1 PBWbNUvJyWuaP3b rFR3cdA2aRxYzbF
MynkiSaYpiXUTEB mt7DC1heGwGB8Re xUpRMwsUtAziUVy 2PrhKtHpECzLAv7
cV8cFYURoDhWCbJ ZyqaENGVtBbauhR jCUVX6qzG4BCBAM nVtgnjBPdz7peGn
WBeXGbistzuFJv7 y41RzPGG71VAfZP hSZ5i8HU4L9KQWi 41f1zuKFdW5qaVZ
mzAsAYHEbtWuLNk CHBNPdbHqBirATA BhwzozweqzwJxj5 FA2wYS5LostE7hq
Y1gHcjLxa2kWPsi ELKNuGxW2pLemKo oYK1aJbpMhVPJaQ GCFvSr56Xejy5Ec
KxKnLBniBFkP7KF 2ZfZ7kF6xjELUdy afkN2cTssWNbfpo fNjBMCkf6jFWXyg
duXPYVJK9KD9U8W wfH1fQpE7zZPohK vonJsgoaFvFc4VT Rp1DgSPELmXAHYZ
gAsvxEyZfFdgFbM iApmLSYwuN8FzsF TqXbTeT9BwAb7i5 FD3s1nRZNpii3y4
YsFUpkZq9WgSz91 Mj8WsnKSh5j2dLq 7Paj5mMC7TUA4a7 zfYB2HQChuATizJ
c5X2yA9pVgQGUTf Dshw65T4jpMoU3e jGVBGMWaXBnyX4Y qdDNfLxcnDAHgUC
5geNtGMMft9ieXL kpxdMXkZ33wjGpe EmikWq46SjpyntA 5NWs4jYCArbFrFU
XhEyiQbn1zmze9j fGWttMcsoPkaHkc 881Vzx1srFsV6h4 siDGJugfFKucX5i
cfdvmHbvx5Sf5Kn cfGmTi6CnQJNT6y 8aSaWdUo8XGrsz. ENDSLATEPACK.
'''

ab_s1 = '''
BEGINSLATEPACK. AHPuvBgs9jCYy5E KAfMU8gTZGFtsPy y9qicNQpkNaiXko
evHryZXtdcuzAnP Jzwkktz2ErCbG9b 8BR6VqLvczzHYWG L6tdBEE1GjqssUH
foRSrRiSoHmimH2 EUisvcn9zb375xx 1b7WNZSYTNgDvfm rmwxeyNE4Q2pVaM
HXi1BzvLj2njAU1 dUxqrT6Fnqkju6P 6RnCNu7UwySae4N RTeUQuyCjv1tZhp
epoFTv9LWn5VtsZ 6xBRVG23AapSWav mr38mS6f3hTB3Ac 9aYbnaFjrBK6jc8
F4a8zvUEfN7gbK1 vC81aVFRUJRzRcz ZSvgzW6DZnJ9tfp jwiTkUNUC9BrtNG
BMx72z6sEKC61jQ RCNMYMLVyFkgc38 iokxKVmLYzVHn9m qJGVUNHitSHo5kP
gvojLmyAvxwFngV ri9BBA9Uqui5ZQz nqSJUzrUv5nYmVj br8Z8nvG4XWLydZ
WZxuWUHFG2Rxhnm iiRpFC1sVhTx2cZ j1QSpzsNror43j1 8vqa3neimQyQu4p
i9DHkAVb7xAKxTy Nh5mt9GcsoR5sTC iqHhKQUowhX6577 nDJpvfueRh8VJuK
EfLjZwhbQR4K2He 5bruawb22wxVHxQ T1fwrDHuz4qptN5 Q8yhmcxbsNL2gEp
piS9m7pfupVEeDc aJKr5YKPk8WcYn5 6CjV6z2XXi7pKDj PCXzrhRzHCqLjBK
N2epP35N1cAnBxm 1cA9Kc3NG5TMVBg bpwYsKQXvFHheGe vMox2WM8WEZY8bT
jqsAKHS. ENDSLATEPACK.
'''

ab_s2 = '''
BEGINSLATEPACK. CE8TL6eVNw1ukDc wgV1EEF5gfgMJhx HiWwgYnKV5xFfdF
fwFPVzzwtqTKqoV wZSgMSc1nVnVYZB D1ZKCYPwHSAA3R3 oNeX4qzhKbJtJUL
vpr8sQ2oJEzXDCk GheGQC9Z3bZ7R56 KDdH2tym4g5L9mi HoagXZPryELYn4E
7ym7YLKZj3oxLU5 bvn9WPVzyPeXL8g mt2YrAZBJ6dJW2y YeQHTHrVSquK2ov
s2G7FFhoCLzbKJd LprxppZCaV6TAdE 8Qbuttb45MeWuKs YsqiaTaQUWK4mSV
ARdvUQ9RdzNgeok topuyehfxZ2u2t2 CW6HemB53vs8Dg8 wtASFn7zsVjh6PC
LXrSN6g3e81KykY Vva2mU7zUWUBoef QFnNpUAGY4Ez7pW PvCHduh1oXxmFCv
7SLH7jpLEqUCc3A zNTDHKTRgXj9VpU Puyakzd7zhKT3jS JbU1ukZp3vm7jwz
vV2hahQQQjnyLkx hLB9LjcwjdPpWBL eMfj4i4xo1vtNLu xCHPYJhWDXDfdj3
k6zRqPtAQKCtke9 8zjnaLiuub2gB1F XgGjrwdZzBzw6eX TCHxgCdKaECnLeS
1NwSsKmwks8cyr7 NbVtWvnckwN1tzS 5PUdMMuWLYRdmrH LZsX1mA9kEEftcj
dhKMuV2eivid5Jx xAo8zFb9NQxEmK6 ESqsfcTrP7r96oV iwwMEvafBFLsH7h
p5VRwxLMLoAqSBg FTawaq8x2RNiHNV NNvzC15VX22G7Cv v3KP3YHsEi4ubiC
fjzKoUhA8sDno4G tKe2nY7UckBj4Ur WZWKjbVXfPTggVK k8i4YtFo4tMM36b
5AgbHZqU66tzyeq j2Y7Nei7NGMf1Lo nCEBWjBFPD3vcW4 5U4gCe7skhZN2AT
tMpeJz2me6RDvTy JJm8bJZbg8i8Ghg 2PtmJJ6sqBodx4s uxvLEgjKo3DbY2M
9QjFu2Zw4Vrm17W ww4rbdQfBiby4MZ ZMQSorVBnsTjHeL wifrbjwyhHtxHX6
VXFVKk2Cwv6K3qZ p3zu7bdyifHeMti UBgGmePBYbky2K2 nyPmRy3nTM1JV5L
dG5pG84JH4Ux3Cv eTMCX38Ah1Q8kjn 876phh1aDH5HX8u UX2GCcwBUABJdCC
W8CTCPRQ6Gq6FGd 5eWNAJ5aZUT63k7 1U6tyBegJiWvVyj rqnfWBEzd1CAwrJ
vzkyoNYWZouZjL3 p56u3nHUfu4fYg3 31rQ3UtLhxDRFq1 5z4oLhFEnEX6vo2
h72Za9ErqGWYTGd f3UgcpnNPoqFwP2 X31j6EPxBYEwgzT dD5fCuHF1bWb5bX
Z5v6UUKSE249WzA bjEBeB1jiZ15qY4 dYeL3hBzB5hSxsx 4dVBc7mi9kBg62c
1bAqoRdR8jkQGAR P5QVwSCC15kpQCG SjRq6wR6DdiuAij NyPHHGDxWDK873T
TRX7pCfum8eLFry 9GJVfZe21pzdDF5 V7oFWL3rVHNoK6Z PJy5VNvKXSkX9fp
o7CKo5bQZ2MBEAV DVKSWZ2RcEcutZw NuLAbSyyLBANR5J dWXZ9fyS9tThcZy
tcEPaUc1VqWHoSx YQ5bd4ijdrgQa1Y SNKCUotRg8tJhCd Pr84rZ6BfEjNtK8
VgZazaxqjwWDbuk uLc5gR32yUJhJ4B NvoZ1EkeABgrGhY K5CyCRCEwhvqxUv
iwrSKwm984aAdQ2 2TvfJ56PHuUYWvr ua1WuCQd548ZX62 9FVRxew29XBoMcy
Lu2xmgXtifuV8qH C5E44j4MGfPEc7k ZTp3YnP1SwHJ1Pg 7UHzorq93pQxSPs
bpctSssfjoVjGnb GHNnTNyMtgRBHXQ 3ycqFPQsohdmQqs Yc174wN3twFGyFU
vorm3ueYN1DZ69r Z12gXBJ5Z4ezh1X arSrW6X5muNkqBZ tKUexx3QRwzaerU
DD3aBesCS3fgVvB WggqQavB1bVXQyb anDbFfN5hrEQ7s9 LxFNipBVN9jmbZV
YAcDHDpA. ENDSLATEPACK.
'''

ab_s3 = '''
BEGINSLATEPACK. 6mgFQ6Xw82pHsaQ FCe3ikRVJpvQwbs kAgMTpRdRjzR1K8
9uKBmG9Re13gy6T pTLNhUTibVo9wZW BEwJ5K1E8ujzSNV 8Kd2FSBef4u72xd
qQHc7gU2Vmc6wJs sDt91xfDB929rcB C8kcYZX2csuxXJu cKW28svofnadW7F
fedv6LYFcsP1FBA ZTDk6GjCz8GxpXq rrPNMpKUaUn1zAj 6KsLRY7PfHtRKLS
U53LzNqmMPsJK7M mEPSdpcjPx3pC81 vnToHA9TnzUtiku jfFztnxNh5icdZM
UWgSdaLPnKUFuHd jZeyEHSVdbAK6gV 8QGtEsJ96yi4pZT 2rKNA63xK3MqrFS
eT1umnVN58AgPpq sURaQxBUHzsS9t3 RNbjtEwiNDWVrTQ CEADif1eVtGc1do
y3C6ELg4Wb9qMa3 zRZ3cWJTLWpg72P 1879hXrCTVyxg4i kLVBGi3sWzyrL8S
uVVp5kHQ2tapsWi fvju1V1JxGMrNnR f5ae1gtDCniAoBc KweezcJaGWSJxT5
Ldoo5kDvgh2cJTa gGfJ7tKW7qzsKMA wnerNjmytXKuhAc BWkuA6dVy5yJb7y
RYhjGb9nt3Dsxkv ffxCMiKSvMLxSLq rpkKcsxoTpYDTEr Hzmr7z28EJyWCXp
DfQs1KHaMaR7Vjy krtSbGQKpw8NoUz UuQwrtSeBU7NNhi 2eKvKSXS1kG1uwz
Y4aDgVYGXvbJwvK n5e1GbVH2Cu8qhH DoAyHVCBXC6LVU2 Xws67RD17y8QKBt
BKsNDFCAdHWbS6z KbdBpWH5aSWWeTe ShfD6qm1riUCjmE feHcbEPT1KA3ZJV
4ZomRxKjExCFcYB CfSU2aTXMVyczEx FjLHT59nySUViLm MzkDkxA55G8URmh
NhEVz1BwoYHQzdM F49ntkGJQw55Wnu z9Hd54NrBDSVCWT H7S34SukJvWyDF8
FEMNv3bvZyBJL9j FbQmwyJi6inso6T f3KspAQFYP5rAH5 QXAgMHAF3g1eJP8
2KEoCr25N8ujW6J BBzHcjPZXp5AeXJ VtZHaAstrzvb28m QpvrLsYwQEqrt7E
ryd2ffU4unZ3EG4 PkguaPSwhx1qCBY tcbUGkQx1Az1HRR oVCfc8HR1N4QAko
67Tve1XfvJQM5aT 3pYHCJXA7RbEq8i 1ikHhBZBS9idU5A EaVESetd4fPuxxa
7MRaz6dtq8XG9Lo XD7ADiZ5yB7EAbL L64gRyT85zdBasT YnzkP2EhsMx67Vy
RjysYwPid2wmbxP W1dTqRqZrG3Qj7j xkHg1jvCLqb4WNx qDMMKbzqcxiwHHM
8ngkUv451mtn5aG RtgiFqiYzgZNR9z TBYLiHoH6zRGn4r ye3uUPsxVxDsuZT
J9ZehQZ2dvzu17M 5JKSuXrXTiCzrms 583UvAvzx22jaUy P2f6ZvsPLraBZJK
1vodaEecLJHag8q FGaqLrBPNXwRcot EFXpFP5qNSVjRtV iRhHEbrKAbFmMmr
zSKjWkD17AQGJQ7 Nof1hK3cCyKFPxg Qb3RFFFJoBfXR8j H6ENFYs5t9bFZMd
4GPThSf6tbNyaRL TzFZVAHtwM6qFY5 VKXQtpb7anRGPLS F9zDrvsfHQJJ6p1
2c9ysqf4CMtzV6Z tWecYj9rYRnkaaH 9UQRe4q8DLa6VP9 5hP2ngNRJ32EYp4
aYkfvy1vL6myHRU troRtT6xR18d2dh GKHfxnNafYwvRZZ g7nojM1rR93JM1W
HPdFpjmHmEEvFQh 5cnjKCooj4SCfVG UcisyT3rJbAdaG. ENDSLATEPACK.
'''


def test_slatepack_manual_decryption():
    # prepare wallets
    w_alice = Wallet.fromSeedPhrase(seed_a)

    # manual way tutorial
    unpacked = SlatepackMessage.unpack(ba_s1)
    s = SlatepackMessage.deserialize(unpacked)

    assert s.toJSON() == {
        'slatepack': [1, 0],
        'mode': 1,
        'payload': 'YWdlLWVuY3J5cHRpb24ub3JnL3YxCi0+IFgyNTUxOSBGL0gzaGY2M0hycTJNSCtPMW5hV202K004MTg0dTNTN29lSFE0VkhpTXpZCnBiVWNMTEhkVDVQR24vb1llaUQrdHNVTzFvZXh1clNabVNWdThIY3Y3bWMKLT4gRXNbLWdyZWFzZSApdksxZUMgbFZ4dVNgLDYKU2ttZ3lDREcwd1BiMHd2R2VZcDZYYmRTeU1PTmZYWW9SUEx5UVNsa3NlOHREeGtUSEt3am96NlkzdDJJSXFIdgpVL2R4eVcyaVp5WjFVVkZScmxOQVZiWGJFQndxaHBhRmpEZFJjaXZoYkRvMXVUTDVDRE52Sno4OXcraXZhb1VjCgotLS0gbnEwaEp2L2JWM2tMUnJxS1U5d2xrZXhGMlNUWkRtQmYwaGxlZXMvSHlXdwqMQcMrGq8Qk1QYD/lKFK+wzXe5EKeQPlYtwz1Er3dj3m8I73jdGYVEdEetyIgqhOVaB0eVCK3KuIFOGkr8l3qo/FLYWaSQzFsFmDmxndFV3HNKdcfEuZEDBwhD1fs+c4AiCYUpjCgdTJ+i6q7yZvh+/sjIFN5LRNiLMtLK9L1Nz+hfiAMkPeERKDVn3pxGLFPsMCMqLqDmSA+UiwLR2XbquHUlRHC0uYhKdnq5Bx4kD/WA5bzlSYufyOEE7DBUmfFjQ9WQsjcYHYidheoHEcSSDAfuxbgf8ibyOQBS1La/ZRj93RqP8MMk/DDz3GDt0L2qvQmd8ZMupqZDDws7nnJ+UL3h1fUf2j595UqTyN70Xp+9u2xL8+NoUO/csr0F8Qie4JauZ9kbx/IvuL4lYVnfxlQ='
    }

    if s.is_encrypted():
        encrypted_payload = s.getPayload()
        decrypted_payload = w_alice.ageDecrypt(
            encrypted_payload, path='m/0/1/0')
        s.setPayload(decrypted_payload, EMode.PLAINTEXT)

    assert s.toJSON() == {
        'slatepack': [1, 0],
        'mode': 0,
        'payload': 'AAQAA1P+Bc0Hi0uPlgtWr28aM7EBAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAGAAAAADrcDeAAAAAAAL68IAEAAlMz9n6EXttW4wnWnBvFa4lU3gmCuizkIRMLchUZp6eaAjLZtJkEPfs1zpdSOzmmDRsiyi6XMmJwS3Wz1h4F1JbpAt1sOfZO8VQjJJ9vPFlCF3tjYUluC2anE8kkBpdXMnu7rZFue/Q2l5j0fYLiGhRWqNeRTAriNW3sgUqboR9t9rIA',
        'sender': 'grin1m4krnajw792zxfyldu79jssh0d3kzjtwpdn2wy7fysrfw4ej0waskurq76'
    }

def test_slatepack_exchange_encryption():
    # prepare wallets
    w_alice = Wallet.fromSeedPhrase(seed_a)
    assert w_alice.getSlatepackAddress(path='m/0/1/0') == address_a

    w_bob = Wallet.fromSeedPhrase(seed_b.replace('\n', ''))
    assert w_bob.getSlatepackAddress(path='m/0/1/0') == address_b

    # Alice decrypts ba_s1 slatepack
    encrypted_slate_message = SlatepackMessage.unarmor(ba_s1)
    assert encrypted_slate_message.is_encrypted()
    assert encrypted_slate_message.toJSON() == {
        'slatepack': [1, 0],
        'mode': 1,
        'payload': 'YWdlLWVuY3J5cHRpb24ub3JnL3YxCi0+IFgyNTUxOSBGL0gzaGY2M0hycTJNSCtPMW5hV202K004MTg0dTNTN29lSFE0VkhpTXpZCnBiVWNMTEhkVDVQR24vb1llaUQrdHNVTzFvZXh1clNabVNWdThIY3Y3bWMKLT4gRXNbLWdyZWFzZSApdksxZUMgbFZ4dVNgLDYKU2ttZ3lDREcwd1BiMHd2R2VZcDZYYmRTeU1PTmZYWW9SUEx5UVNsa3NlOHREeGtUSEt3am96NlkzdDJJSXFIdgpVL2R4eVcyaVp5WjFVVkZScmxOQVZiWGJFQndxaHBhRmpEZFJjaXZoYkRvMXVUTDVDRE52Sno4OXcraXZhb1VjCgotLS0gbnEwaEp2L2JWM2tMUnJxS1U5d2xrZXhGMlNUWkRtQmYwaGxlZXMvSHlXdwqMQcMrGq8Qk1QYD/lKFK+wzXe5EKeQPlYtwz1Er3dj3m8I73jdGYVEdEetyIgqhOVaB0eVCK3KuIFOGkr8l3qo/FLYWaSQzFsFmDmxndFV3HNKdcfEuZEDBwhD1fs+c4AiCYUpjCgdTJ+i6q7yZvh+/sjIFN5LRNiLMtLK9L1Nz+hfiAMkPeERKDVn3pxGLFPsMCMqLqDmSA+UiwLR2XbquHUlRHC0uYhKdnq5Bx4kD/WA5bzlSYufyOEE7DBUmfFjQ9WQsjcYHYidheoHEcSSDAfuxbgf8ibyOQBS1La/ZRj93RqP8MMk/DDz3GDt0L2qvQmd8ZMupqZDDws7nnJ+UL3h1fUf2j595UqTyN70Xp+9u2xL8+NoUO/csr0F8Qie4JauZ9kbx/IvuL4lYVnfxlQ='
    }

    decrypted_slate_message = w_alice.decryptSlatepack(ba_s1, path='m/0/1/0')

    assert decrypted_slate_message.toJSON() == {
        'slatepack': [1, 0],
        'mode': 0,
        'payload': 'AAQAA1P+Bc0Hi0uPlgtWr28aM7EBAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAGAAAAADrcDeAAAAAAAL68IAEAAlMz9n6EXttW4wnWnBvFa4lU3gmCuizkIRMLchUZp6eaAjLZtJkEPfs1zpdSOzmmDRsiyi6XMmJwS3Wz1h4F1JbpAt1sOfZO8VQjJJ9vPFlCF3tjYUluC2anE8kkBpdXMnu7rZFue/Q2l5j0fYLiGhRWqNeRTAriNW3sgUqboR9t9rIA',
        'sender': 'grin1m4krnajw792zxfyldu79jssh0d3kzjtwpdn2wy7fysrfw4ej0waskurq76'
    }

    assert decrypted_slate_message.getSlate().toJSON() == {
        'ver': '4:3',
        'id': '53fe05cd-078b-4b8f-960b-56af6f1a33b1',
        'sta': 'S1',
        'amt': 987500000,
        'fee': 12500000,
        'sigs': [
            {
                'xs': '025333f67e845edb56e309d69c1bc56b8954de0982ba2ce421130b721519a7a79a',
                'nonce': '0232d9b499043dfb35ce97523b39a60d1b22ca2e973262704b75b3d61e05d496e9'
            }
        ],
        'proof': {
            'saddr': 'dd6c39f64ef15423249f6f3c5942177b6361496e0b66a713c924069757327bbb',
            'raddr': 'ad916e7bf4369798f47d82e21a1456a8d7914c0ae2356dec814a9ba11f6df6b2'
        }
    }

    # Bob decrypts ba_s2 slatepack
    encrypted_slate_message = SlatepackMessage.unarmor(ba_s2)
    assert encrypted_slate_message.is_encrypted()
    assert encrypted_slate_message.toJSON() == {
        'slatepack': [1, 0],
        'mode': 1,
        'payload': 'YWdlLWVuY3J5cHRpb24ub3JnL3YxCi0+IFgyNTUxOSBmbVFzbFlUeURuK3R1SUd3WmtiYk9PTUd0djY3c3lrWVYzaXd0OUh6UjNnCkp6eDFhR1YvZ1orN2FkRDdHWmRqN2JsN3RlNTFYVUZINVZJMzM2RWI3SUEKLT4gOCZoMkNLLWdyZWFzZSBVSiFQaDQwPwpYZUlJWkdNV3QraDBmQ3pXR0RUVnUvM29Kc2N2WWoxbGRFNmhlSnEwa0MxbmhMOGhBd1Z3bVhQbFdjcGNPais5CjBMdVRPMTRDCi0tLSBrdFE1YUpvLzBhaWNYcGh0dGw3YmpTK0t5ejVQbllYd1lZRjJqeVFsUjJ3CjBzNyM8pnZbCELsjLt2MrUUxJG7Fm665i/y24utTxzsPzdzhBpVash8Kgxys5e1XtBYJEOqcWmoxX0BZg1BAKezfqT5rNJ3NEhju0J029gZ3oKYd9tYE2ly5r3ZtiXOyEsvwJ8oxA7GN+UuQ5TYnLgDZnV4QVUv+Ho/vOXoqNFkc7a4J3PgtpNg3EGeB0rlMHkwMkKneNMPm/IrdGhnREe8mUMPR9tFRZiYsuVnQ65Yuuc2R3OrJ0Sw45wTxMCnt3MtazX5MC0s10wUXqz5xbwvi86t7mJhZdoeC0aGZt/Zy/lh05sPfQyWndHF8HPzdBWfxavjOGK1vPcuwynzfGj7klpMuqwixMiR8svuJ9ITY9NdJMWHVtH7DplOMybUEVYIAzJ8qp8721VflHo9PAP3Mpd896z5JM9tlPPnCbql+ZvEdW3x/eC604vUyijTMc/ODOdW9u8Kyn4vfstFkh5godmWCtZeQ5lMpcZb3EKdZs4/1AcYQM5uhZyBAI9ugbF/Yvo42xVlDUMV733cyD44rlnOPwd9ZnGjUOwoiOm7Knz6saNc4hm1k07AkERhYLVirpznVkIrVt3IEzDpPbPUsdD0T8vKcSBsDJgb3S7XmlXPIocn9jDmH64PBfh6KJlE+Smfzd/1NOX94km6YMxMQJJ4gptGRp99bNTBCVaXUX6qPxf57pE00hrr8mVYdSD+dZUX3CTqCPmjxS3mKjVYzBA3w9tqCoS1Ynamcv9XgHGQ16hbymuLBbP0PrD+f07TiCYUaeQeVkiAZ0lcPDXtGA1zsjkBEANAST1+wsXLzYuUkGGQHbbsrWGY/DKfp36vkr2fM2TuFadLHdQPOaV/RqYmbW+NE1pyzhNaHw6xsrlvmQV2irnsleA+N7sOkweN3H0RQDQOdCiD7HEnAr7kxsEpUnk+nsgiLQuBf6sbL4OQ2BAxqjZvfQdlCgPSh0UvrcK2FHJvdzzh6YvaDSttvQgzljypHhO18dfp+S3ely33IqAOQuo+5deoejMafEBWaCIVU0jPIHPAFzYc9b+itcHeqYiXpBpaVVcOBscSRL+CG48PsV19GLPpQj1SLqN5E+8IquIGlhYvWsFB2IU4CJrLHF6Ohk1g2jT6bVf2YS/C77XWYWalJAoAadyW1SaaNRfQ8/WzDFoHa2PNV+Kg5VRnhfD7i53eq+WJedXBOSnhP2b1VYHGnmYBHngiTOwzuWl3ehvJyVO9d9YXYYdlKp197KjUo3igbUoswvmLNVycT1UhZyeFN6h36fzoCJvy0N04j6ScF9ndbuBauvnzWUYv7SkNp+X+twSQXdIIlAaPEHTlw67AlEgQ75nkCOJfnBDO+Cd9fVaECdYMT5n3QvYEm2vORaXlDgwPYtY+xoJqbe5zCY3HduZ1N6+0tE1kObfUOysXwf6DWBh3nec4m4tHNgYpYhTF4GCjXlXX9zComLKIkitb8kGxzRBM+AsErw6S9vvpr3zPInLr/htx6s5Pst5dJM9DoYjN/QZV50k='
    }

    decrypted_slate_message = w_bob.decryptSlatepack(ba_s2, path='m/0/1/0')

    assert decrypted_slate_message.toJSON() == {
        'slatepack': [1, 0],
        'mode': 0,
        'payload': 'AAQAA1P+Bc0Hi0uPlgtWr28aM7ECQEiMwcqE4OJ0v4WrC58Sq8wYoIyD6f6jFCWrA2uMo48AAQEDCwq0jluOoUy+IeAG4DY/sUmpIHl9DUgGDfUNhsxDAisD+9oFyODFaa5IRF+Xq/+7L0wrdnlq5Xtk4O8Nb5cHAaL72gXI4MVprkhEX5er/7svTCt2eWrle2Tg7w1vlwcBoig1KMskAAB69AIK+2MBdSWpSIcw6220+mMa+lajJUzyAwABAQAJb/ypAPNhBVfE9Qc+zBiClBWBE6qIFhVWIYP18samsmoAAAAAAAACo2RHJ5usYpRb4hEPP+srfIlzRPMrJVyBsPerZMy9Vxfy1SCnJaojLzXUT/rbWh7/01u8Z+j+lTLfLFCWgePCne4AUmgXcz6qjPBZ9D75gOgnjQ4vptkyh4iBunGycMOT4arXJjf4ZpBPq+O+Ri+8tix1Fc71vXvJMILMpZ+OhUCWQ6IcLKHni0o5ENqHt0iZueY/LYqfD+benJ/T7ZA3gLphasWLU4DRscsKUNi/SaJeb+Bs//VWmRwgwdLg/Q63TnQfMP0Pi++K7cbgd3mK9//2xqL8MvOHJeovC1iYn+4gsb036MebrBhe5gqg3BdxCEw9I0us3iU4ONGHj3f9y0Hxt/ZjtkPC2hfRwl9FptUGe7hMEkkV1faHL5+whZnbq9Tgkt3E6CRiuuXhDPk/wD12M1aBIfIvUKw53ELfmW4Sy04/JV0hfj6JYCPdLvbmYR4bCYJW+OGJOxJU74CBTLyT+gAzzJ9yp/VpllfJuNxXlwiEYFz03BqdtV8ZhhQmD88LAJRfF9pDTek1WNCDctl928aFYCY/xRNWxjtaG4FysyrMVNq380Rsp5WwDUCqRHdJshFPz+HnOZN31yo4aTSwQe49El59d7Tn03dYdtN3pU4BxmuJDuDR7prHryB6MSVVo7ABCnyACilmSvFmOp8y1vgCTbJXVrTj0MVARmv/O/yg5XA5CS59BnxKcAkKaTvCs6LZvXRb03kTNaDamycmvKzI23bCZTTDCwDBU1u9yB4+KLITF+pKx5TWr+6L91SSV585ClVwDVUV5zWkOZm1uSUGPI9fY0UnHN6yzwiKWCAtJdigaiV6FRBJfwz49fCeUCsu2OA1IUNQJ0dA4WKfZvyKhK2sgO3ny+sIHQG6U1LiSAmcpuVo/Ka2mpRAso8sM91sOfZO8VQjJJ9vPFlCF3tjYUluC2anE8kkBpdXMnu7rZFue/Q2l5j0fYLiGhRWqNeRTAriNW3sgUqboR9t9rIBjGWkBYvACKTsep3bSE9bV/9AHteumW4ldwZCx9hpTPCucZGwYDIDlonWPIrIPpWRHCUjRFERkvyF2AhsNJFECQ==',
        'sender': 'grin14kgku7l5x6te3arast3p59zk4rteznq2ug6kmmypf2d6z8md76eqg3su35'
    }

    assert decrypted_slate_message.getSlate().toJSON() == {
        'ver': '4:3',
        'id': '53fe05cd-078b-4b8f-960b-56af6f1a33b1',
        'sta': 'S2',
        'off': '40488cc1ca84e0e274bf85ab0b9f12abcc18a08c83e9fea31425ab036b8ca38f',
        'sigs': [
            {
                'xs': '030b0ab48e5b8ea14cbe21e006e0363fb149a920797d0d48060df50d86cc43022b',
                'nonce': '03fbda05c8e0c569ae48445f97abffbb2f4c2b76796ae57b64e0ef0d6f970701a2',
                'part': 'a20107976f0defe0647be56a79762b4c2fbbffab975f4448ae69c5e0c805dafbf24c25a356fa1a63fab46deb308748a925750163fb0a02f47a000024cb283528'
            }
        ],
        'coms': [
            {
                'c': '096ffca900f3610557c4f5073ecc188294158113aa881615562183f5f2c6a6b26a',
                'p': '6447279bac62945be2110f3feb2b7c897344f32b255c81b0f7ab64ccbd5717f2d520a725aa232f35d44ffadb5a1effd35bbc67e8fe9532df2c509681e3c29dee00526817733eaa8cf059f43ef980e8278d0e2fa6d932878881ba71b270c393e1aad72637f866904fabe3be462fbcb62c7515cef5bd7bc93082cca59f8e85409643a21c2ca1e78b4a3910da87b74899b9e63f2d8a9f0fe6de9c9fd3ed903780ba616ac58b5380d1b1cb0a50d8bf49a25e6fe06cfff556991c20c1d2e0fd0eb74e741f30fd0f8bef8aedc6e077798af7fff6c6a2fc32f38725ea2f0b58989fee20b1bd37e8c79bac185ee60aa0dc1771084c3d234bacde253838d1878f77fdcb41f1b7f663b643c2da17d1c25f45a6d5067bb84c124915d5f6872f9fb08599dbabd4e092ddc4e82462bae5e10cf93fc03d7633568121f22f50ac39dc42df996e12cb4e3f255d217e3e896023dd2ef6e6611e1b098256f8e1893b1254ef80814cbc93fa0033cc9f72a7f5699657c9b8dc57970884605cf4dc1a9db55f198614260fcf0b00945f17da434de93558d08372d97ddbc68560263fc51356c63b5a1b8172b32acc54dab7f3446ca795b00d40aa447749b2114fcfe1e7399377d72a386934b041ee3d125e7d77b4e7d3775876d377a54e01c66b890ee0d1ee9ac7af207a312555a3b0010a7c800a29664af1663a9f32d6f8024db25756b4e3d0c540466bff3bfca0e57039092e7d067c4a70090a693bc2b3a2d9bd745bd3791335a0da9b2726bcacc8db76c26534c30b00c1535bbdc81e3e28b21317ea4ac794d6afee8bf75492579f390a55700d5515e735a43999b5b925063c8f5f6345271cdeb2cf088a58202d25d8a06a257a1510497f0cf8f5f09e502b2ed8e035214350274740e1629f66fc8a84adac80ede7cbeb081d01ba5352e248099ca6e568fca6b69a9440b28f2c33'
            }
        ],
        'proof': {
            'saddr': 'dd6c39f64ef15423249f6f3c5942177b6361496e0b66a713c924069757327bbb',
            'raddr': 'ad916e7bf4369798f47d82e21a1456a8d7914c0ae2356dec814a9ba11f6df6b2',
            'rsig': '8c65a4058bc008a4ec7a9ddb484f5b57ff401ed7ae996e25770642c7d8694cf0ae7191b06032039689d63c8ac83e95911c252344511192fc85d8086c34914409'
        }
    }


    # Bob finalizes and makes the ba_s3 slatepack
    decrypted_slate_message = SlatepackMessage.unarmor(ba_s3)
    assert not decrypted_slate_message.is_encrypted()

    assert decrypted_slate_message.toJSON() == {
        'slatepack': [1, 0],
        'mode': 0,
        'sender': 'grin1m4krnajw792zxfyldu79jssh0d3kzjtwpdn2wy7fysrfw4ej0waskurq76',
        'payload': 'AAQAA1P+Bc0Hi0uPlgtWr28aM7EDIeyFsJzBL7KTpA+x8TNiWWuyP3prlZUI28ZAqFB2Xd4EAAAAAAC+vCACAQMLCrSOW46hTL4h4AbgNj+xSakgeX0NSAYN9Q2GzEMCKwP72gXI4MVprkhEX5er/7svTCt2eWrle2Tg7w1vlwcBovvaBcjgxWmuSERfl6v/uy9MK3Z5auV7ZODvDW+XBwGiKDUoyyQAAHr0Agr7YwF1JalIhzDrbbT6Yxr6VqMlTPIBAlMz9n6EXttW4wnWnBvFa4lU3gmCuizkIRMLchUZp6eaAjLZtJkEPfs1zpdSOzmmDRsiyi6XMmJwS3Wz1h4F1JbpMtm0mQQ9+zXOl1I7OaYNGyLKLpcyYnBLdbPWHgXUluk7wfBwJgPlSW7xw3ka1oM3q74mc3dyfy1+D9K2Vc1ynQMAAgAACCqQq7/DLLzi5Jbz46qzyBZx77KicvXZGGxUqttsj8OFAQAJb/ypAPNhBVfE9Qc+zBiClBWBE6qIFhVWIYP18samsmoAAAAAAAACo2RHJ5usYpRb4hEPP+srfIlzRPMrJVyBsPerZMy9Vxfy1SCnJaojLzXUT/rbWh7/01u8Z+j+lTLfLFCWgePCne4AUmgXcz6qjPBZ9D75gOgnjQ4vptkyh4iBunGycMOT4arXJjf4ZpBPq+O+Ri+8tix1Fc71vXvJMILMpZ+OhUCWQ6IcLKHni0o5ENqHt0iZueY/LYqfD+benJ/T7ZA3gLphasWLU4DRscsKUNi/SaJeb+Bs//VWmRwgwdLg/Q63TnQfMP0Pi++K7cbgd3mK9//2xqL8MvOHJeovC1iYn+4gsb036MebrBhe5gqg3BdxCEw9I0us3iU4ONGHj3f9y0Hxt/ZjtkPC2hfRwl9FptUGe7hMEkkV1faHL5+whZnbq9Tgkt3E6CRiuuXhDPk/wD12M1aBIfIvUKw53ELfmW4Sy04/JV0hfj6JYCPdLvbmYR4bCYJW+OGJOxJU74CBTLyT+gAzzJ9yp/VpllfJuNxXlwiEYFz03BqdtV8ZhhQmD88LAJRfF9pDTek1WNCDctl928aFYCY/xRNWxjtaG4FysyrMVNq380Rsp5WwDUCqRHdJshFPz+HnOZN31yo4aTSwQe49El59d7Tn03dYdtN3pU4BxmuJDuDR7prHryB6MSVVo7ABCnyACilmSvFmOp8y1vgCTbJXVrTj0MVARmv/O/yg5XA5CS59BnxKcAkKaTvCs6LZvXRb03kTNaDamycmvKzI23bCZTTDCwDBU1u9yB4+KLITF+pKx5TWr+6L91SSV585ClVwDVUV5zWkOZm1uSUGPI9fY0UnHN6yzwiKWCAtJdigaiV6FRBJfwz49fCeUCsu2OA1IUNQJ0dA4WKfZvyKhK2sgO3ny+sIHQG6U1LiSAmcpuVo/Ka2mpRAso8sM91sOfZO8VQjJJ9vPFlCF3tjYUluC2anE8kkBpdXMnu7rZFue/Q2l5j0fYLiGhRWqNeRTAriNW3sgUqboR9t9rIBjGWkBYvACKTsep3bSE9bV/9AHteumW4ldwZCx9hpTPCucZGwYDIDlonWPIrIPpWRHCUjRFERkvyF2AhsNJFECQ=='
    }

    assert decrypted_slate_message.getSlate().toJSON() == {
        'ver': '4:3',
        'id': '53fe05cd-078b-4b8f-960b-56af6f1a33b1',
        'sta': 'S3',
        'off': '21ec85b09cc12fb293a40fb1f13362596bb23f7a6b959508dbc640a850765dde',
        'fee': 12500000,
        'sigs': [
            {
                'xs': '030b0ab48e5b8ea14cbe21e006e0363fb149a920797d0d48060df50d86cc43022b',
                'nonce': '03fbda05c8e0c569ae48445f97abffbb2f4c2b76796ae57b64e0ef0d6f970701a2',
                'part': 'a20107976f0defe0647be56a79762b4c2fbbffab975f4448ae69c5e0c805dafbf24c25a356fa1a63fab46deb308748a925750163fb0a02f47a000024cb283528'
            },
            {
                'xs': '025333f67e845edb56e309d69c1bc56b8954de0982ba2ce421130b721519a7a79a',
                'nonce': '0232d9b499043dfb35ce97523b39a60d1b22ca2e973262704b75b3d61e05d496e9',
                'part': 'e996d4051ed6b3754b706232972eca221b0da6393b5297ce35fb3d0499b4d9329d72cd55b6d20f7e2d7f72777326beab3783d61a79c3f16e49e5032670f0c13b'
            }
        ],
        'coms': [
            {
                'c': '082a90abbfc32cbce2e496f3e3aab3c81671efb2a272f5d9186c54aadb6c8fc385'
            },
            {
                'c': '096ffca900f3610557c4f5073ecc188294158113aa881615562183f5f2c6a6b26a',
                'p': '6447279bac62945be2110f3feb2b7c897344f32b255c81b0f7ab64ccbd5717f2d520a725aa232f35d44ffadb5a1effd35bbc67e8fe9532df2c509681e3c29dee00526817733eaa8cf059f43ef980e8278d0e2fa6d932878881ba71b270c393e1aad72637f866904fabe3be462fbcb62c7515cef5bd7bc93082cca59f8e85409643a21c2ca1e78b4a3910da87b74899b9e63f2d8a9f0fe6de9c9fd3ed903780ba616ac58b5380d1b1cb0a50d8bf49a25e6fe06cfff556991c20c1d2e0fd0eb74e741f30fd0f8bef8aedc6e077798af7fff6c6a2fc32f38725ea2f0b58989fee20b1bd37e8c79bac185ee60aa0dc1771084c3d234bacde253838d1878f77fdcb41f1b7f663b643c2da17d1c25f45a6d5067bb84c124915d5f6872f9fb08599dbabd4e092ddc4e82462bae5e10cf93fc03d7633568121f22f50ac39dc42df996e12cb4e3f255d217e3e896023dd2ef6e6611e1b098256f8e1893b1254ef80814cbc93fa0033cc9f72a7f5699657c9b8dc57970884605cf4dc1a9db55f198614260fcf0b00945f17da434de93558d08372d97ddbc68560263fc51356c63b5a1b8172b32acc54dab7f3446ca795b00d40aa447749b2114fcfe1e7399377d72a386934b041ee3d125e7d77b4e7d3775876d377a54e01c66b890ee0d1ee9ac7af207a312555a3b0010a7c800a29664af1663a9f32d6f8024db25756b4e3d0c540466bff3bfca0e57039092e7d067c4a70090a693bc2b3a2d9bd745bd3791335a0da9b2726bcacc8db76c26534c30b00c1535bbdc81e3e28b21317ea4ac794d6afee8bf75492579f390a55700d5515e735a43999b5b925063c8f5f6345271cdeb2cf088a58202d25d8a06a257a1510497f0cf8f5f09e502b2ed8e035214350274740e1629f66fc8a84adac80ede7cbeb081d01ba5352e248099ca6e568fca6b69a9440b28f2c33'
            }
        ],
        'proof': {
            'saddr': 'dd6c39f64ef15423249f6f3c5942177b6361496e0b66a713c924069757327bbb',
            'raddr': 'ad916e7bf4369798f47d82e21a1456a8d7914c0ae2356dec814a9ba11f6df6b2',
            'rsig': '8c65a4058bc008a4ec7a9ddb484f5b57ff401ed7ae996e25770642c7d8694cf0ae7191b06032039689d63c8ac83e95911c252344511192fc85d8086c34914409'
        }
    }






from mimblewimble.models.slatepack.address import SlatepackAddress
from mimblewimble.models.slatepack.metadata import SlatepackVersion
from mimblewimble.models.slatepack.metadata import SlatepackMetadata
from mimblewimble.models.slatepack.message import SlatepackMessage, EMode

sender = 'grin1m3ckuft0llw97ulzmedk557twyn76wkgxartn6lpfq5xd9wcc24qnpwde5'

recipients = [
    'grin1at3z7s2fa796zcqurhyjyklhpfx0rgn9czynshv9rq0yq674wkdq5qdy6w',
    'grin18r9pen0fjemhss2tf92dkmm8puaflw22vlretwmukq7tefdqgaeqyve0qy',
    'grin1qaq69t9pugrh9h2vamc9qhc09vplaa2zy2ukzwcd08236j5qgzwsmasmf5'
]

payload_hex = '7513a0cafc4a87ecef0144c3ca93f9f5a2e4a3cdc143993029525c060ae80904'


def test_slatepack_message():
    version = SlatepackVersion(0, 0)
    metadata = SlatepackMetadata(
        sender=SlatepackAddress.fromBech32(sender),
        recipients=[
            SlatepackAddress.fromBech32(recipient) for recipient in recipients
        ])
    emode = EMode.ENCRYPTED
    payload = bytes.fromhex(payload_hex)
    slatepack_message = SlatepackMessage(
        version, metadata, emode, payload)

    assert slatepack_message.is_encrypted()
    assert slatepack_message.toJSON() == {
        'slatepack': [0, 0],
        'mode': 1,
        'payload': 'dROgyvxKh+zvAUTDypP59aLko83BQ5kwKVJcBgroCQQ='
    }





from mimblewimble.helpers.tor import TorAddress
from mimblewimble.keychain import KeyChain

# test data from https://github.com/GrinPlusPlus/GrinPlusPlus/blob/fa3705f5558f0372410aa4222f5f1d97a0ab2247/tests/src/Wallet/Test_Slatepack.cpp

test_data = [
    (
        'm/0/1/0',
        'eae22f4149ef8ba1601c1dc9225bf70a4cf1a265c089385d85181e406bd5759a',
        '5lrc6qkj56f2cya4dxesew7xbjgpditfycetqxmfdapea26vownfpnid.onion',
        'grin1at3z7s2fa796zcqurhyjyklhpfx0rgn9czynshv9rq0yq674wkdq5qdy6w'
    ),
    (
        'm/0/1/1',
        '38ca1ccde9967778414b4954db6f670f3a9fb94a67c795bb7cb03cbca5a04772',
        'hdfbztpjsz3xqqkljfknw33hb45j7okkm7dzlo34wa6lzjnai5zj25yd.onion',
        'grin18r9pen0fjemhss2tf92dkmm8puaflw22vlretwmukq7tefdqgaeqyve0qy'
    ),
    (
        'm/0/1/2',
        '0741a2aca1e20772dd4ceef0505f0f2b03fef54222b9613b0d79d51d4a80409d',
        'a5a2flfb4idxfxkm53yfaxypfmb755kcek4wcoynphkr2suaicotbyad.onion',
        'grin1qaq69t9pugrh9h2vamc9qhc09vplaa2zy2ukzwcd08236j5qgzwsmasmf5'
    ),
    (
        'm/0/1/3',
        '897bdd8906a2e6ed5e0b545f5edb765502f48d4d8ce130ad4fa307425d7b4ded',
        'rf553cigulto2xqlkrpv5w3wkubpjdknrtqtblkpumduexl3jxw5v2qd.onion',
        'grin139aamzgx5tnw6hst2304akmk25p0fr2d3nsnpt205vr5yhtmfhksd4wg9s'
    ),
    (
        'm/0/1/4',
        'dc716e256fffdc5f73e2de5b6a53cb7127ed3ac83746b9ebe148286695d8c2aa',
        '3ryw4jlp77of647c3znwuu6loet62owig5dlt27bjaugnfoyykvneayd.onion',
        'grin1m3ckuft0llw97ulzmedk557twyn76wkgxartn6lpfq5xd9wcc24qnpwde5'
    )
]

def test_slatepack_address():
    seed = bytes.fromhex(
        '09e626b6322a9459ed1dd6bbfcacda2e4d3558d7a1631576ff104e04fb387710')
    keychain = KeyChain.fromSeed(seed)

    for path, expected_ed25519, expected_onion, expected_slatepack_address in test_data:
        ed25519 = keychain.deriveED25519PublicKey(path).hex()
        onion = keychain.deriveOnionAddress(path)
        slatepack_address = keychain.deriveSlatepackAddress(path)
        assert ed25519 == expected_ed25519
        assert onion == expected_onion
        assert slatepack_address == expected_slatepack_address

        # validate the TOR address
        TorAddress.parse(onion)



import pytest

from tests.utils import MockNode

from mimblewimble.wallet import Wallet
from mimblewimble.wallet import WalletStorageInMemory, PersistentWallet, NodeAccess

from mimblewimble.models.slatepack.address import SlatepackAddress
from mimblewimble.models.slatepack.metadata import SlatepackVersion
from mimblewimble.models.slatepack.metadata import SlatepackMetadata
from mimblewimble.models.slatepack.message import SlatepackMessage, EMode


def test_rsr_flow_slatepacks():
    # prepare wallets
    alice_path = 'm/0/1/0'
    alice_wallet = Wallet.initialize()
    alice_address = alice_wallet.getSlatepackAddress()

    bob_path = 'm/0/1/0'
    bob_wallet = Wallet.initialize()
    bob_address = bob_wallet.getSlatepackAddress()

    # prepare transaction data
    coinbase_amount = 60000000000
    kernel, output = alice_wallet.createCoinbase(
        coinbase_amount, path=alice_path)

    num_change_outputs = 1
    amount = 30000000000
    fee_base = 7000000
    block_height = 99999

    # prepare invoice transaction
    invoice_slate, bob_secret_key, bob_secret_nonce = bob_wallet.invoice(amount, path=bob_path, block_height=block_height)
    invoice_slate_payload = invoice_slate.serialize()

    # prepare R1 invoice slatepack
    version = SlatepackVersion(0, 0)
    metadata = SlatepackMetadata(
        sender=SlatepackAddress.fromBech32(alice_address))
    emode = EMode.PLAINTEXT
    invoice_slatepack_message = SlatepackMessage(
        version, metadata, emode, invoice_slate_payload)
    assert not invoice_slatepack_message.is_encrypted()

    # encrypt the R1 slatepack
    invoice_slatepack_message.encryptPayload(
        [SlatepackAddress.fromBech32(alice_address)])
    slatepack_R1 = invoice_slatepack_message.pack()
    assert invoice_slatepack_message.is_encrypted()
    # now Bob sends slatepack_R1 to Alice

    # Alice receives it
    received_slatepack_R1 = alice_wallet.decryptSlatepack(
        slatepack_R1, path=alice_path)
    assert received_slatepack_R1.payload == invoice_slate_payload
    received_invoice_slate = received_slatepack_R1.getSlate()

    # now Alice has decrypted slate that Bob sent her
    # she is preparing her own slate
    pay_slate, alice_secret_key, alice_secret_nonce, change_outputs = alice_wallet.pay(
        received_invoice_slate,
        [output],
        num_change_outputs,
        fee_base,
        alice_path)
    pay_slate_payload = pay_slate.serialize()

    # prepare slatepack_R2 for Bob
    version = SlatepackVersion(0, 0)
    metadata = SlatepackMetadata(
        sender=SlatepackAddress.fromBech32(alice_address))
    emode = EMode.PLAINTEXT
    pay_slatepack_message = SlatepackMessage(
        version, metadata, emode, pay_slate_payload)
    assert not pay_slatepack_message.is_encrypted()

    pay_slatepack_message.encryptPayload(
        [SlatepackAddress.fromBech32(bob_address)])
    slatepack_R2 = pay_slatepack_message.pack()
    # now Alice sends slatepack_R2 back to Bob

    # Bob receive it
    received_pay_slatepack_message = bob_wallet.decryptSlatepack(
        slatepack_R2, path=bob_path)
    assert received_pay_slatepack_message.payload == pay_slate_payload
    received_pay_slate = received_pay_slatepack_message.getSlate()

    # now Bob has decrypted slate that Alice sent him
    # he is ready to finalize the transaction
    finalized_slate = bob_wallet.finalize(
        received_pay_slate, path=bob_path)
    # this is the final slate that can be broadcasted to nodes


def test_rsr_flow_slatepacks_persistent():
    mock_node = MockNode()

    # prepare wallets
    alice_path = 'm/0/1/0'
    alice_wallet = PersistentWallet(
        Wallet.initialize(), WalletStorageInMemory(), mock_node)
    alice_recovery_phrase = alice_wallet.getSeedPhrase()

    bob_path = 'm/0/1/0'
    bob_wallet = PersistentWallet(
        Wallet.initialize(), WalletStorageInMemory(), mock_node)
    bob_recovery_phrase = bob_wallet.getSeedPhrase()

    # transaction setup
    fee_base = 7000000

    # check balance, every wallet is empty
    assert alice_wallet.balance().toJSON() == {
        'locked': 0,
        'awaiting_confirmation': 0,
        'spendable': 0,
        'total': 0
    }
    assert bob_wallet.balance().toJSON() == {
        'locked': 0,
        'awaiting_confirmation': 0,
        'spendable': 0,
        'total': 0
    }

    # Bob prepares invoice transaction
    r1_slatepack_message = bob_wallet.invoice(
        30000000000,
        path=bob_path,
        receiver=alice_wallet.getSlatepackAddress())
    r1_slatepack_text = r1_slatepack_message.pack()  # this text gets sent back to Bob

    # now Alice receives R1 slatepack
    r1_slatepack_message_received = SlatepackMessage.unarmor(r1_slatepack_text)
    assert r1_slatepack_message_received.is_encrypted()

    # Alice attempts to pay the invoice, but has no funds yet
    with pytest.raises(Exception, match='Insufficient funds'):
        alice_wallet.pay(
            r1_slatepack_message, path=alice_path, fee_base=fee_base)

    # Alice needs some coinbase to pay Bob's invoice
    coinbase_amount = 60000000000
    coinbase_transaction = alice_wallet.createCoinbase(
        coinbase_amount, path=alice_path)

    # submit the coinbase transaction to the node
    mock_node.push_transaction(coinbase_transaction)

    # mine a block to confirm Alice's coinbase
    mock_node.mine()

    # now Alice needs to refresh her wallet to detect confirmed funds
    alice_wallet.refresh()

    # Alice should have her confirmed coinbase balance now
    assert alice_wallet.balance().toJSON() == {
        'locked': 0,
        'awaiting_confirmation': 0,
        'spendable': 60000000000,
        'total': 60000000000
    }

    # Alice pays the invoice
    r2_slatepack_message = alice_wallet.pay(
        r1_slatepack_message, path=alice_path, fee_base=fee_base)
    r2_slatepack_text = r2_slatepack_message.pack()  # this text gets sent back to Bob

    # extract slate from R2 slatepack to check fee
    # we use Bob's wallet as Alice encrypted the slatepack for Bob
    # unlike in SRS, in RSR flow the fee gets decided at Pay step (second step)
    slate_fee = bob_wallet.decryptSlatepack(r2_slatepack_message).getSlate().getFee().getFee()
    assert slate_fee == 322000000

    # now Bob receives R2 slatepack
    r2_slatepack_message_received = SlatepackMessage.unarmor(r2_slatepack_text)
    assert r2_slatepack_message_received.is_encrypted()

    # they both check their wallets to see locked output amounts
    alice_balance = alice_wallet.balance().toJSON()
    assert alice_balance == {
        'locked': 60000000000,
        'awaiting_confirmation': 30000000000 - slate_fee,
        'spendable': 0,
        'total': 30000000000 - slate_fee,
    }

    bob_balance = bob_wallet.balance().toJSON()
    assert bob_balance == {
        'locked': 0,
        'awaiting_confirmation': 30000000000,
        'spendable': 0,
        'total': 30000000000
    }

    # Bob finalizes the transaction
    finalized_slate = bob_wallet.finalize(
        r2_slatepack_message, path=bob_path)

    # Bob sends transaction to node
    bob_wallet.push_finalized_slatepack(finalized_slate)

    # mine a block to confirm the transaction
    mock_node.mine()

    # they both check their wallets
    alice_wallet.refresh()
    bob_wallet.refresh()

    # check their balances
    alice_balance = alice_wallet.balance().toJSON()
    assert alice_balance == {
        'locked': 0,
        'awaiting_confirmation': 0,
        'spendable': 30000000000 - slate_fee,
        'total': 30000000000 - slate_fee,
    }

    bob_balance = bob_wallet.balance().toJSON()
    assert bob_balance == {
        'locked': 0,
        'awaiting_confirmation': 0,
        'spendable': 30000000000,
        'total': 30000000000
    }

    # test restoring wallet from seed phrase and syncing with node
    # first Alice
    alice_wallet_restored = PersistentWallet.restoreFromSeedPhrase(
        alice_recovery_phrase, WalletStorageInMemory(), mock_node)
    assert not alice_wallet_restored.is_synced()

    alice_wallet_restored.scan()

    alice_balance_restored = alice_wallet_restored.balance().toJSON()
    assert alice_balance_restored == {
        'locked': 0,
        'awaiting_confirmation': 0,
        'spendable': 30000000000 - slate_fee,
        'total': 30000000000 - slate_fee,
    }
    assert alice_wallet_restored.is_synced()

    # now Bob
    bob_wallet_restored = PersistentWallet.restoreFromSeedPhrase(
        bob_recovery_phrase, WalletStorageInMemory(), mock_node)
    assert not bob_wallet_restored.is_synced()

    bob_wallet_restored.scan()

    bob_balance_restored = bob_wallet.balance().toJSON()
    assert bob_balance_restored == {
        'locked': 0,
        'awaiting_confirmation': 0,
        'spendable': 30000000000,
        'total': 30000000000
    }

    assert bob_wallet_restored.is_synced()


import os
import pytest

from mimblewimble.keychain import KeyChain

from mimblewimble.crypto.commitment import Commitment
from mimblewimble.crypto.bulletproof import EBulletproofType
from mimblewimble.crypto.pedersen import Pedersen

from mimblewimble.models.transaction import BlindingFactor


def test_rewind_bulletproof_original():
    master_seed = os.urandom(32)

    bulletproof_type = EBulletproofType.ORIGINAL

    path = 'm/0/1/0'
    amount = 45
    keychain = KeyChain.fromSeed(master_seed)
    blinding_factor = keychain.derivePrivateKeyAmount(
        path, amount)

    p = Pedersen()
    commitment = p.commit(amount, blinding_factor)
    del p

    rangeproof = keychain.generateRangeProof(
        path, amount, commitment,
        blinding_factor, bulletproof_type)

    rewoundproof = keychain.rewindRangeProof(
        commitment, rangeproof, bulletproof_type)

    assert amount == rewoundproof.getAmount()
    assert blinding_factor.getBytes() == rewoundproof.getBlindingFactor().getBytes()
    assert KeyChain.getKeyIndices(path) == rewoundproof.toKeyIndices(bulletproof_type)

def test_rewind_bulletproof_enhanced():
    master_seed = os.urandom(32)

    bulletproof_type = EBulletproofType.ENHANCED

    path = 'm/0/1/0'
    amount = 45
    keychain = KeyChain.fromSeed(master_seed)
    blinding_factor = keychain.derivePrivateKeyAmount(
        path, amount)

    p = Pedersen()
    commitment = p.commit(amount, blinding_factor)
    del p

    rangeproof = keychain.generateRangeProof(
        path, amount, commitment,
        blinding_factor, bulletproof_type)

    rewoundproof = keychain.rewindRangeProof(
        commitment, rangeproof, bulletproof_type)

    assert amount == rewoundproof.getAmount()
    assert KeyChain.getKeyIndices(path) == rewoundproof.toKeyIndices(bulletproof_type)



import hmac
import os

from typing import Union

from bip32 import BIP32
from bip_utils import Bech32Encoder
from hashlib import blake2b, pbkdf2_hmac, sha512

from nacl import bindings
from nacl.signing import SigningKey, VerifyKey

from mimblewimble.crypto.bulletproof import EBulletproofType

from mimblewimble.crypto.commitment import Commitment
from mimblewimble.crypto.secret_key import SecretKey
from mimblewimble.crypto.public_keys import PublicKeys
from mimblewimble.crypto.rangeproof import RangeProof
from mimblewimble.crypto.pedersen import Pedersen

from mimblewimble.crypto.bulletproof import EBulletproofType
from mimblewimble.crypto.bulletproof import ProofMessage
from mimblewimble.crypto.bulletproof import RewoundProof
from mimblewimble.crypto.bulletproof import Bulletproof

from mimblewimble.models.slatepack.address import SlatepackAddress

from mimblewimble.helpers.encryption import ageX25519Decrypt


class KeyChain:
    def __init__(self, master_key: bytes, bulletproof_nonce: SecretKey):
        self.master_key = master_key
        self.bulletproof_nonce = bulletproof_nonce

    def derivePrivateKey(self, path: str):
        # derive the seed at the path
        master_key = self.master_key.getBytes()
        bip32 = BIP32(
            chaincode=master_key[32:], privkey=master_key[:32])
        sk = bip32.get_privkey_from_path(path)
        return SecretKey(sk)

    def derivePrivateKeyAmount(self, path: str, amount: int):
        p = Pedersen()
        sk = self.derivePrivateKey(path)
        ska = p.blindSwitch(sk, amount)
        return ska

    def deriveED25519Seed(self, path: str):
        sk_der = self.derivePrivateKey(path)

        # compute the blake2 hash of that key and that is ed25519 seed
        seed_blake = blake2b(sk_der.getBytes(), digest_size=32).digest()
        return seed_blake

    def deriveED25519SecretKey(self, path: str):
        # get the seed
        seed_blake = self.deriveED25519Seed(path)

        # get the ed25519 secret key and public key from it
        pk, sk = bindings.crypto_sign_seed_keypair(seed_blake)
        return sk

    def deriveED25519PublicKey(self, path: str):
        # get the seed
        seed_blake = self.deriveED25519Seed(path)

        # get the ed25519 public key and public key from it
        pk, sk = bindings.crypto_sign_seed_keypair(seed_blake)
        return pk

    def deriveX25519PublicKey(self, path: str):
        ed25519_pk = self.deriveED25519PublicKey(path=path)

        x25519_pk = bindings.crypto_sign_ed25519_pk_to_curve25519(ed25519_pk)

        return x25519_pk

    def deriveX25519SecretKey(self, path: str):
        # get the seed
        # seed_blake = self.deriveED25519Seed(path)

        # get the ed25519 public key and public key from it
        # ed25519_pk, ed25519_sk = bindings.crypto_sign_seed_keypair(seed_blake)

        ed25519_sk = self.deriveED25519SecretKey(path=path)

        # construct the x25519 secret key
        x25519_sk = bindings.crypto_sign_ed25519_sk_to_curve25519(ed25519_sk)

        return x25519_sk

    def deriveAgeSecretKey(self, path: str):
        x25519_sk = self.deriveX25519SecretKey(path)

        age_secret_key = Bech32Encoder.Encode('age-secret-key-', x25519_sk)
        return age_secret_key.upper()

    def signED25519(self, message: bytes, path: str) -> bytes:
        sk = self.deriveED25519SecretKey(path)
        signature_message = bindings.crypto_sign(message, sk)
        del sk
        # first 64 bytes is signature, then message, this is just how nacl works
        # but grin only needs signature
        signature = signature_message[0:64]
        return signature

    @classmethod
    def verifyED25519(
            self,
            public_key: Union[bytes, str],
            signature: bytes,
            message: bytes) -> bool:
        public_key_bytes = public_key
        if isinstance(public_key, str):
            public_key_bytes = KeyChain.slatepackAddressToED25519PublicKey(
                public_key)
        signature_message = signature + message
        recovered = bindings.crypto_sign_open(signature_message, public_key_bytes)
        return recovered == message

    def ageDecrypt(
            self,
            ciphertext: bytes,
            path: str):
        age_secret_key = self.deriveAgeSecretKey(path)

        return ageX25519Decrypt(
            ciphertext, age_secret_key)

    def deriveSlatepackAddress(self, path: str, testnet=False):
        pk = self.deriveED25519PublicKey(path)
        slatepack_address = SlatepackAddress(pk)
        return slatepack_address.toBech32(testnet=testnet)

    def deriveOnionAddress(self, path: str):
        pk = self.deriveED25519PublicKey(path)
        return KeyChain.slatepackAddressToOnion(pk)

    @classmethod
    def slatepackAddressToED25519PublicKey(
            self, address: str, testnet=False) -> bytes:
        slatepack_address = SlatepackAddress.fromBech32(
            address, testnet=testnet)
        return slatepack_address.toED25519()

    @classmethod
    def slatepackAddressToOnion(self, public_key: Union[bytes, str]):
        slatepack_address = SlatepackAddress(public_key)
        if isinstance(public_key, str):
            slatepack_address = SlatepackAddress.fromBech32(
                public_key)
        return slatepack_address.toOnion(version=3)

    def rewindRangeProof(
            self,
            commitment: Commitment,
            rangeproof: RangeProof,
            bulletproof_type: EBulletproofType):
        b = Bulletproof()
        try:
            if bulletproof_type == EBulletproofType.ORIGINAL:
                nonce = KeyChain.createNonce(
                    commitment, self.bulletproof_nonce)
                rewind_result = b.rewindProof(
                    commitment, rangeproof, nonce)
            elif bulletproof_type == EBulletproofType.ENHANCED:
                pks = PublicKeys()
                master_public_key = pks.calculatePublicKey(self.master_key)

                rewind_nonce_hash = SecretKey(blake2b(
                    master_public_key.getBytes(),
                    digest_size=32).digest())

                rewind_result = b.rewindProof(
                    commitment,
                    rangeproof,
                    KeyChain.createNonce(commitment, rewind_nonce_hash))
            return rewind_result
        finally:
            del b  # just to make sure secp256k1-zkp context is freed
        raise ValueError('Unimplemented bulletproof type')

    def generateRangeProof(
            self,
            path: str,
            amount: int,
            commitment: Commitment,
            blinding_factor: SecretKey,
            bulletproof_type: EBulletproofType):

        key_indices = KeyChain.getKeyIndices(path)
        proof_message = ProofMessage.fromKeyIndices(
            key_indices, bulletproof_type)

        b = Bulletproof()
        if bulletproof_type == EBulletproofType.ORIGINAL:
            nonce = KeyChain.createNonce(
                commitment, self.bulletproof_nonce)
            generated_range_proof = b.generateRangeProof(
                amount, blinding_factor, nonce, nonce, proof_message)
            del b # just to make sure secp256k1-zkp context is freed
            return generated_range_proof
        elif bulletproof_type == EBulletproofType.ENHANCED:
            private_nonce_hash = SecretKey(blake2b(
                self.master_key.getBytes(),
                digest_size=32).digest())

            pks = PublicKeys()
            master_public_key = pks.calculatePublicKey(self.master_key)

            rewind_nonce_hash = SecretKey(blake2b(
                master_public_key.getBytes(),
                digest_size=32).digest())

            generated_range_proof = b.generateRangeProof(
                amount,
                blinding_factor,
                KeyChain.createNonce(commitment, private_nonce_hash),
                KeyChain.createNonce(commitment, rewind_nonce_hash),
                proof_message)
            del b # just to make sure secp256k1-zkp context is freed
            return generated_range_proof
        del b # just to make sure secp256k1-zkp context is freed
        raise ValueError('Unimplemented bulletproof type')

    @classmethod
    def getKeyIndices(self, path: str):
        key_indices = []
        for v in path.split('/'):
            try:
                i = int(v)
            except:
                continue
            key_indices.append(i)
        return key_indices

    @classmethod
    def createNonce(self, commitment: Commitment, nonce_hash: SecretKey):
        nonce = blake2b(
            commitment.getBytes() + nonce_hash.getBytes(),
            digest_size=32).digest()
        return SecretKey(nonce)

    @classmethod
    def fromSeed(self, master_seed):
        # I AM VOLDEMORT
        m = hmac.new('IamVoldemort'.encode('utf8'), digestmod=sha512)
        m.update(master_seed)
        master_key = SecretKey(m.digest())
        p = Pedersen()
        bulletproof_nonce = p.blindSwitch(master_key, 0)
        return KeyChain(master_key, bulletproof_nonce)

    @classmethod
    def fromRandom(self):
        master_seed = os.urandom(32)
        return KeyChain.fromSeed(master_seed)



import os

from nacl import bindings

from mimblewimble.keychain import KeyChain

def test_nacl():
    seed = os.urandom(32)
    pk, sk = bindings.crypto_sign_seed_keypair(seed)

    message = os.urandom(128)
    signature = bindings.crypto_sign(message, sk)

    msg = bindings.crypto_sign_open(signature, pk)
    assert msg == message


def test_signature():
    master_seed = os.urandom(32)
    path = 'm/0/1/0'

    keychain = KeyChain.fromSeed(master_seed)

    message = os.urandom(128)
    signature = keychain.signED25519(message, path)
    address = keychain.deriveSlatepackAddress(path)

    valid = KeyChain.verifyED25519(address, signature, message)
    assert valid



import pytest

from io import BytesIO

import hashlib
import unittest

from mimblewimble.serializer import Serializer
from mimblewimble.genesis import floonet, mainnet
from mimblewimble.blockchain import FullBlock


class GenesisTest(unittest.TestCase):
    def test_mainnet(self):
        genesis = mainnet
        # check proof of work hash
        assert genesis.getHash().hex() == '40adad0aec27797b48840aa9e00472015c21baea118ce7a2ff1a82c0f8f5bf82'
        # check serialized block header hash
        serializer = Serializer()
        genesis.header.serialize(serializer)
        computed = hashlib.blake2b(serializer.getvalue(), digest_size=32).digest()
        computed == bytes.fromhex('6046b08903fa520e693b86e700d6a430b353e767027dd9ed1ccbb7017d0bb4dc') # value extracted from grin++
        # check whole serialized block hash
        computed = hashlib.blake2b(genesis.serialize(), digest_size=32).digest()
        assert computed == bytes.fromhex('6be6f34b657b785e558e85cc3b8bdb5bcbe8c10e7e58524c8027da7727e189ef')

    def test_floonet(self):
        genesis = floonet
        # check proof of work hash
        assert genesis.getHash().hex() == 'edc758c1370d43e1d733f70f58cf187c3be8242830429b1676b89fd91ccf2dab'
        # check serialized block header hash
        serializer = Serializer()
        genesis.header.serialize(serializer)
        computed = hashlib.blake2b(serializer.getvalue(), digest_size=32).digest()
        computed == bytes.fromhex('2d71d7996a0cf2eda1c357eb11e713e1ef7930de5d293a839c4e4d40e37d3e1d') # value extracted from grin++
        # check whole serialized block hash
        computed = hashlib.blake2b(genesis.serialize(), digest_size=32).digest()
        assert computed == bytes.fromhex('91c638fc019a54e6652bd6bb3d9c5e0c17e889cef34a5c28528e7eb61a884dc4')

    def test_mainnet_deserialize(self):
        genesis = mainnet
        serialized = genesis.serialize()
        deserializer = Serializer()
        deserializer.write(serialized)
        deserialized = FullBlock.deserialize(deserializer)
        hash_serialized = hashlib.blake2b(serialized, digest_size=32).digest()
        hash_deserialized = hashlib.blake2b(deserialized.serialize(), digest_size=32).digest()
        assert hash_serialized == hash_deserialized

    def test_floonet_deserialize(self):
        genesis = floonet
        serialized = genesis.serialize()
        deserializer = Serializer()
        deserializer.write(serialized)
        deserialized = FullBlock.deserialize(deserializer)
        hash_serialized = hashlib.blake2b(serialized, digest_size=32).digest()
        hash_deserialized = hashlib.blake2b(deserialized.serialize(), digest_size=32).digest()
        assert hash_serialized == hash_deserialized



from io import BytesIO
from enum import Enum

class EProtocolVersion(Enum):
    V1 = 1
    V2 = 2
    V3 = 3
    V1000 = 1000

class Serializer:
    def __init__(self, protocol=EProtocolVersion.V1):
        self.pnt = 0
        self.protocol = protocol
        self.raw = BytesIO()

    def resetPointer(self, n=None):
        if n is None:
            self.pnt = 0
        else:
            self.pnt = n

    def read(self, n):
        self.raw.seek(self.pnt)
        self.pnt += n
        return self.raw.read(n)

    def write(self, b: bytes):
        self.raw.write(b)

    def readline(self, clean_newline=False):
        self.raw.seek(self.pnt)
        line = self.raw.readline()
        self.pnt = self.raw.tell()
        if clean_newline:
            if line[-1] == ord('\n'):
                return line[0:-1]
        return line

    def readlines(self):
        lines = []
        line = self.readline()
        while line != b'':
            lines.append(line)
            line = self.readline()
        if line != b'':
            lines.append(line)
        return lines

    def readString(self, n, encoding='ascii'):
        as_bytes = self.read(n)
        return as_bytes.decode(encoding)

    def writeString(self, value: str, encoding='ascii'):
        as_bytes = value.encode(encoding)
        self.write(as_bytes)

    def getvalue(self):
        return self.raw.getvalue()

    def getProtocol(self):
        return self.protocol

    def readremaining(self):
        self.raw.seek(self.pnt)
        value = self.raw.read()
        self.pnt = self.raw.tell()
        return value

    def readall(self):
        self.raw.seek(0)
        value = self.raw.read()
        self.raw.seek(self.pnt)
        return value

    def getStream(self):
        return self.raw

    def __len__(self):
        return self.raw.getbuffer().nbytes



import math


class Consensus:
    # A grin is divisible to 10^9, following the SI prefixes
    grin_base = 1000000000

    # Milligrin, a thousand of a grin
    milli_grin = grin_base/1000

    # Microgrin, a thousand of a milligrin
    micro_grin = milli_grin/1000

    # Nanogrin, smallest unit, takes a billion to make a grin
    nano_grin = 1

    # The block subsidy amount, one grin per second on average
    reward = 60*grin_base

    # Weight of an input when counted against the max block weight capacity
    input_weight = 1

    # Weight of an output when counted against the max block weight capacity
    output_weight = 21

    # Weight of a kernel when counted against the max block weight capacity
    kernel_weight = 3

    # Total maximum block weight
    max_block_weight = 40000

    @classmethod
    def calculateWeightV4(self, num_inputs, num_outputs, num_kernels):
        return -1*num_inputs + 4*num_outputs + num_kernels

    @classmethod
    def calculateWeightV5(self, num_inputs, num_outputs, num_kernels):
        return num_inputs*Consensus.input_weight + num_outputs*Consensus.output_weight + num_kernels*Consensus.kernel_weight

    # Block interval, in seconds
    block_time_sec = 60

    # Nominal height for standards time intervals, hour is 60 blocks
    hour_height = 3600 / block_time_sec

    @classmethod
    def hours(self, num_hours):
        return num_hours*Consensus.hour_height

    # A day is 1440 blocks
    day_height = 24*hour_height

    @classmethod
    def days(self, num_days):
        return num_days*Consensus.day_height

    # A week is 10,080 blocks
    week_height = 7*day_height

    @classmethod
    def weeks(self, num_weeks):
        return num_weeks*Consensus.week_height

    # A year is 524,160 blocks
    year_height = 52*week_height

    @classmethod
    def years(self, num_years):
        return num_years*Consensus.year_height

    # Number of blocks before a coinbase matures and can be spent
    # set to nominal number of block in one day (1440 with 1-minute blocks)
    coinbase_maturity = (24*60*60)/block_time_sec

    @classmethod
    def getMaxCoinbaseHeight(blockHeight, automated_testing=False):
        if automated_testing:
            return math.max(blockHeight, 25)-20
        return math.max(blockHeight, Consensus.coinbase_maturity)-Consensus.coinbase_maturity

    # Default number of blocks in the past when cross-block cut-through will start happening
    cut_through_horizon = week_height

    @classmethod
    def getHorizonHeight(block_height):
        return math.max(block_height, Consensus.cut_through_horizon) - Consensus.cut_through_horizon

    # Default number of blocks in the past to determine the height where we request a txhashset
    state_sync_threshold = 2*day_height

    # Time window in blocks to calculate block time median
    median_time_window = 11

    # Index at half the desired median
    median_time_index = median_time_window/2

    # Number of blocks used to calculate difficulty adjustments
    difficulty_adjust_window = hour_height

    # Average time span of the difficulty adjustment window
    block_time_window = difficulty_adjust_window*block_time_sec

    # Maximum size time window used for difficulty adjustments
    upper_time_bound = block_time_window*2

    # Minimum size time window used for difficulty adjustments
    lower_time_bound = block_time_window/2

    # default Future Time Limit (FTL) of 5 minutes
    default_future_time_limit_sec = 5*block_time_sec

    # Refuse blocks more than 12 block intervals in the future.
    @classmethod
    def getMaxBlockTime(current_time):
        return Consensus.currentTime + Consensus.default_future_time_limit_sec

    # Difficulty adjustment half life (actually, 60s * number of 0s-blocks to raise diff by factor e) is 4 hours
    wtema_half_life = 4*3600

    # Cuckoo-cycle proof size (cycle length)
    proofsize = 42

    # Default Cuckoo Cycle size shift used for mining and validating.
    default_min_edge_bits = 31

    # Secondary proof-of-work size shift, meant to be ASIC resistant.
    second_pow_edge_bits = 29

    # Original reference edge_bits to compute difficulty factors for higher Cuckoo graph sizes, changing this would hard fork
    base_edge_bits = 24

    # Clamp factor to use for difficulty adjustment
	# Limit value to within this factor of goal
    clamp_factor = 2

    # Dampening factor to use for difficulty adjustment
    dma_damp_factor = 3

    # Dampening factor to use for AR scale calculation.
    ar_scale_damp_factor = 13

    # Minimum scaling factor for AR pow, enforced in diff retargetting
	# avoids getting stuck when trying to increase ar_scale subject to dampening
    min_ar_scale = ar_scale_damp_factor

    # Minimum difficulty, enforced in diff retargetting
	# avoids getting stuck when trying to increase difficulty subject to dampening
    min_dma_difficulty = dma_damp_factor

    # Compute weight of a graph as number of siphash bits defining the graph
    # Must be made dependent on height to phase out smaller size over the years
    @classmethod
    def graphWeight(self, height, edge_bits):
        expiry_height = int(Consensus.year_height)
        xpr_edge_bits = int(edge_bits)
        if edge_bits == 31 and height >= expiry_height:
            xpr_edge_bits -= math.min(xpr_edge_bits, 1+(height-expiry_height)/Consensus.week_height)
        return 2 << (edge_bits-Consensus.base_edge_bits)*xpr_edge_bits

    # Initial mining secondary scale
    @classmethod
    def initialGraphWeight(self):
        return Consensus.graphWeight(0, Consensus.second_pow_edge_bits)

    # Move value linearly toward a goal
    @classmethod
    def damp(actual, goal, damp_factor):
        return (actual + (damp_factor-1))/damp_factor

    # limit value to be within some factor from a goal
    @classmethod
    def clamp(actual, goal, clamp_factor):
        return math.max(goal/clamp_factor, math.min(actual, goal*clamp_factor))

    # Ratio the secondary proof of work should take over the primary, as a function of block height (time).
    # Starts at 90% losing a percent approximately every week. Represented as an integer between 0 and 100.
    @classmethod
    def secondaryPOWRatio(self, height):
        return 90-math.min(90, (height/(2*Consensus.year_height/90)))

    @classmethod
    def scalingDifficulty(self, edgeBits):
        # TODO find a pure python way to ensure "2" is of uint64_t type
        return 2 << (edgeBits-Consensus.base_edge_bits)*edgeBits

    # minimum solution difficulty after HardFork4 when pow becomes primary only Cuckatoo32+
    # TODO find a pure python way to ensure "2" is of uint64_t type
    c32_graph_weight = 2 << (32-base_edge_bits)*32

    def min_wtema_graph_weight(self, testnet=False):
        if testnet:
            return self.GraphWeight(0, Consensus.second_pow_edge_bits)
        return Consensus.c32_graph_weight

    # Fork every 6 months.
    hard_fork_interval = year_height / 2

    # Floonet-only hardforks
    floonet_first_hard_fork = 185040
    floonet_second_hard_fork = 298080
    floonet_third_hard_fork = 552960
    floonet_fourth_hard_fork = 642240

    def getHeaderVersion(height, testnet=False):
        if testnet:
            if height < Consensus.floonet_first_hard_fork:
                return 1
            elif height < Consensus.floonet_second_hard_fork:
                return 2
            elif height < Consensus.floonet_third_hard_fork:
                return 3
            elif height < Consensus.floonet_fourth_hard_fork:
                return 4
        else:
            if height < Consensus.hard_fork_interval:
                return 1
            elif height < 2*Consensus.hard_fork_interval:
                return 2
            elif height < 3*Consensus.hard_fork_interval:
                return 3
            elif height < 4*Consensus.hard_fork_interval:
                return 4
        return 5

    def __init__(self):
        pass

    @classmethod
    def isPrimary(edgeBits):
        pass

    @classmethod
    def isSecondary(edgeBits):
        pass
